"""
Utilities for terminals
"""

import termcolor


class ColorFormatter:
    def __init__(self, theme="dark"):
        if theme is None:
            theme = "none"
        themes = ["dark", "light", "none"]
        assert theme in themes, f"Unknown color theme {theme}, should be any of {themes}"
        self.theme = theme

    def fmt_bold(self, inp):
        if self.theme == "none":
            return inp
        return termcolor.colored(inp, attrs=["bold"])

    def fmt_correct(self, inp):
        if self.theme == "none":
            return inp
        if self.theme == "dark":
            return termcolor.colored(inp, "light_green")
        return termcolor.colored(inp, "green")

    def fmt_wrong(self, inp):
        if self.theme == "none":
            return inp
        if self.theme == "dark":
            return termcolor.colored(inp, "light_red")
        return termcolor.colored(inp, "red")

    def fmt_neutral(self, inp):
        if self.theme == "none":
            return inp
        return inp
        # return termcolor.colored(inp, "white")

    def fmt_underline(self, inp):
        if self.theme == "none":
            return inp
        return termcolor.colored(inp, attrs=["underline"])

    def fmt_grey(self, inp):
        if self.theme == "none":
            return inp
        if self.theme == "dark":
            return termcolor.colored(inp, "light_grey")
        return termcolor.colored(inp, "grey")



"""
Color utilities for matplotlib and jupyter notebook
"""

from dataclasses import dataclass
from matplotlib import colorbar as mpl_colorbar, colors as mpl_colors, pyplot as plt, cm as mpl_cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from typing import Union, Tuple

redgreen_bright = mpl_colors.LinearSegmentedColormap.from_list(
    "rwg", [[1.0, 0.7, 0.7], [1.0, 1.0, 1.0], [0.7, 1.0, 0.7]]
)

redgreen_dark = mpl_colors.LinearSegmentedColormap.from_list(
    "rg", [[0.5, 0, 0], [0, 0, 0], [0, 0.3, 0]]
)


rbg_bright = mpl_colors.LinearSegmentedColormap.from_list(
    "rbg",
    [
        [1.0, 0.7, 0.7],
        [1.0, 1.0, 0.7],
        [1.0, 1.0, 1.0],
        [0.7, 1.0, 1.0],
        [0.7, 1.0, 0.7],
    ],
)

rbg_dark = mpl_colors.LinearSegmentedColormap.from_list(
    "rbg",
    [
        [0.5, 0, 0],
        [0.3, 0.3, 0],
        [0.0, 0.0, 0.0],
        [0.0, 0.3, 0.3],
        [0.0, 0.3, 0.0],
    ],
)


DEFAULT_COLOR_CYCLE = plt.rcParams["axes.prop_cycle"].by_key()["color"]


def get_redgreen_for_theme(is_dark_theme: bool):
    return redgreen_dark if is_dark_theme else redgreen_bright


@dataclass
class ColorGetterForColorMap:
    input_cmap: Union[mpl_colors.Colormap, str]
    start_val: Union[int, float]
    stop_val: Union[int, float]
    return_alpha: bool = False
    return_float: bool = False

    def __post_init__(self):
        self.cmap = self.input_cmap
        if isinstance(self.input_cmap, str):
            self.cmap = plt.get_cmap(self.input_cmap)
        self.norm = mpl_colors.Normalize(vmin=self.start_val, vmax=self.stop_val)
        self.scalar_map = mpl_cm.ScalarMappable(norm=self.norm, cmap=self.cmap)

    def get_color(self, val) -> Union[Tuple[int, int, int, int], Tuple[int, int, int]]:
        rgba = self.scalar_map.to_rgba(val)
        if not self.return_alpha:
            rgba = rgba[:3]
        if not self.return_float:
            rgba = [int(round(a * 255)) for a in rgba]
        return rgba


def create_colorgetter_from_colormap(
    input_cmap: Union[mpl_colors.Colormap, str],
    return_alpha: bool = False,
    return_float: bool = False,
):
    # getter is now function that takes a value in (0, 1) and returns a color
    color_helper = ColorGetterForColorMap(
        input_cmap, 0, 1, return_alpha=return_alpha, return_float=return_float
    )
    return color_helper.get_color


def create_colormap_for_dark_background(
    n: int = 256, inverted=False, scale=1.0, alpha=1.0, add_brightness: float = 0.0
) -> mpl_colors.Colormap:
    """
    Colormap with only light colors that works well for text on dark background.

    Args:
        n: number of discrete colors to generate in the colormap
        inverted: if True, invert the colormap (gives dark colors for bright background)
        scale: downscale the colormap (make it darker, 1.0 = no scaling)
        alpha: transparency value for all colors, 1.0 = opaque, 0.0 = transparent / invisible
        add_brightness: add this to all RGB values (0.0 = no change, 1.0 = all colors are white)

    Returns:
        colormap
    """
    ccmap_colors = [
        [128, 128, 255],
        [0, 255, 255],
        [0, 255, 0],
        [255, 255, 0],
        [255, 0, 0],
        [255, 0, 255],
        [255, 255, 255],
    ]

    def _convert(c):
        c = c / 255
        if inverted:
            c = 1 - c
        c_scaled = max(min(c * scale, 1.0), 0.0)
        return c_scaled * (1.0 - add_brightness) + add_brightness

    ccmap_colors_01 = [[_convert(a) for a in rgb] + [alpha] for rgb in ccmap_colors]
    ccmap = mpl_colors.LinearSegmentedColormap.from_list("better_nipy", ccmap_colors_01, N=n)
    return ccmap


rainbow_bright = create_colormap_for_dark_background()
rainbow_dark = create_colormap_for_dark_background(inverted=True)


def add_colorbar(fig, fig_imshow, ax):
    """Add colorbar to a given figure"""
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(fig_imshow, cax=cax)


def show_colormap(input_cmap: Union[mpl_colors.Colormap, str]):
    """
    Show a colormap in a figure

    Args:
        input_cmap: Colormap or name of colormap

    Returns:

    """
    if isinstance(input_cmap, str):
        input_cmap = plt.get_cmap(input_cmap)
    cmap_norm = mpl_colors.Normalize(vmin=0.0, vmax=1.0)
    fig, ax = plt.subplots(1, 1, figsize=(16, 1))
    mpl_colorbar.ColorbarBase(ax, cmap=input_cmap, norm=cmap_norm, orientation="horizontal")
    fig.tight_layout()
    return fig, ax


def get_color_from_color_cycle(num, cycle):
    return cycle[num % len(cycle)]


def get_color_from_default_color_cycle(num):
    return get_color_from_color_cycle(num, DEFAULT_COLOR_CYCLE)


def get_color_from_simple_transition(i: int, i_max: int):
    """
    g->b->r color transition, 0 <= i < i_max
    """
    fact = (i / (i_max - 1)) * 2
    bb = 0.4
    if fact <= 1:
        cr = fact * bb
        cg = (1 - fact) * (1 - bb) + bb
        cb = fact
    else:
        fact -= 1
        cr = (fact * (1 - bb)) + bb
        cg = (1 - fact) * bb
        cb = 1 - fact
    return cr, cg, cb


def get_listed_colormap_from_default_cycle(alpha: float = 1.0):
    """
    Create a ListedColormap from the default color cycle.
    Use cmap.colors to get the colors back.
    """
    # input is a list of hex colors e.g. ['#1f77b4', ...] length 10
    new_list = DEFAULT_COLOR_CYCLE

    # convert to rgb, add alpha
    rgb_list = [mpl_colors.hex2color(a) for a in new_list]
    rgba_list = [(r, g, b, alpha) for r, g, b in rgb_list]

    cmap_bold = mpl_colors.ListedColormap(rgba_list)
    return cmap_bold



from pathlib import Path
from timeit import default_timer

from visiontext.examples import cc0images
from visiontext.images import JPEGDecoderConst, decode_jpeg

image_shapes_highres = [(5184, 3456, 3), (2038, 3057, 3), (2560, 2048, 3)]
image_shapes_lowres = [(500, 333, 3), (333, 500, 3), (500, 400, 3)]


def main():
    src_dir_base = Path(cc0images.__file__).parent

    for src_dir, image_shapes, n_trials in [
        (src_dir_base, image_shapes_highres, 2),
        (src_dir_base / "lowres", image_shapes_lowres, 100),
    ]:
        print(f"Test images in {src_dir}")
        images_f = list(sorted(src_dir.glob("*.jpg")))

        if len(images_f) != len(image_shapes):
            raise ValueError(
                f"Expected {len(image_shapes)} images in {src_dir}, got {len(images_f)}"
            )
        images_b = [Path(image_f).read_bytes() for image_f in images_f]

        for decoder in JPEGDecoderConst.values():
            if decoder == JPEGDecoderConst.PILLOW_IMAGE:
                continue
            t0 = default_timer()
            for n in range(n_trials):
                for i, image_b in enumerate(images_b):
                    arr = decode_jpeg(image_b, method=decoder)
                    if n == 0:
                        # only test shape for first trial to not disturb the performance test
                        ex_s = image_shapes[i]
                        act_s = arr.shape
                        if ex_s != act_s:
                            raise ValueError(
                                f"Expected shape {ex_s} for {images_f[i]}, got {act_s}"
                            )
            tdelta = default_timer() - t0
            tper_ms = (tdelta / n_trials / len(images_f)) * 1000
            print(f"    Decoder {decoder} took {tper_ms:.2f} ms per image on average")


if __name__ == "__main__":
    main()



from __future__ import annotations

import re

RE_ALNUM = re.compile(r"[^a-zA-Z0-9 ]+")


def preprocess_text_simple(in_str: str, lowercase=True, regexp: re.Pattern | str = RE_ALNUM) -> str:
    """Remove non-alphanumeric characters, lowercase, and strip duplicate whitespaces."""
    if not isinstance(regexp, re.Pattern):
        regexp = re.compile(regexp)
    intermed_str = regexp.sub(" ", in_str).strip()
    if lowercase:
        intermed_str = intermed_str.lower()
    return " ".join(intermed_str.split())



"""
Undo the normalization of an image to go from [-1, 1] back to the original range for display.
"""

import torch
import torchvision
import torchvision.transforms.v2

from packg.log import logger


class Denormalize:
    def __init__(self, mean: torch.Tensor, std: torch.Tensor):
        self.mean = torch.tensor(mean, dtype=torch.float32)
        self.std = torch.tensor(std, dtype=torch.float32)
        # print(f"Created normalizer with shape {self.mean.shape} {self.std.shape} ")

    def __call__(self, image_tensor: torch.Tensor):
        """
        Args:
            image_tensor: either (C, H, W) or (B, C, H, W)

        Returns:
            same image_tensor but with normalization undone, for human viewing

        """
        # print(f"Got input with shape {tensor.shape} {tensor.dtype}")
        # normalize op is: output[channel] = (input[channel] - mean[channel]) / std[channel]
        # therefore denormalize is: output[channel] = input[channel] * std[channel] + mean[channel]
        new_dims = [1] * len(image_tensor.shape)

        if image_tensor.ndim == 4:
            new_dims[1] = -1
        elif image_tensor.ndim == 3:
            new_dims[0] = -1
        else:
            raise ValueError(f"Expected 3 or 4 dims, got shape {image_tensor.shape}")
        mean = self.mean.view(*new_dims).to(image_tensor.device)
        std = self.std.view(*new_dims).to(image_tensor.device)
        # print(f"mean {mean.shape} {mean.dtype} std {std.shape} {std.dtype}")
        return image_tensor * std + mean

    @classmethod
    def from_transforms(cls, transforms, ignore_errors: bool = False):
        mean, std = find_mean_and_std_for_denormalization(transforms)
        if mean is None or std is None:
            err_msg = f"Could not find mean and std for denormalization in {transforms=}"
            if ignore_errors:
                logger.warning(err_msg)
            else:
                raise ValueError(err_msg)
            mean, std = [0.0, 0.0, 0.0], [1.0, 1.0, 1.0]
        return cls(mean, std)

    def __repr__(self):
        return f"Denormalize(mean={self.mean}, std={self.std})"


def find_mean_and_std_for_denormalization(transforms):
    mean, std = None, None

    # support ViTImageProcessor
    if hasattr(transforms, "image_mean") and hasattr(transforms, "image_std"):
        return transforms.image_mean, transforms.image_std

    # call recursively if it is a list of transforms
    transform_list = None
    if isinstance(transforms, list):
        transform_list = transforms
    elif isinstance(
        transforms, (torchvision.transforms.Compose, torchvision.transforms.v2.Compose)
    ):
        transform_list = transforms.transforms
    if transform_list is not None:
        for tf in transform_list:
            mean, std = find_mean_and_std_for_denormalization(tf)
            if mean is not None and std is not None:
                break
        return mean, std

    if hasattr(transforms, "mean") and hasattr(transforms, "std"):
        mean, std = transforms.mean, transforms.std
    return mean, std



import lmdb
import msgpack
import os
from pathlib import Path
from typing import Optional

from packg.paths import get_cache_dir
from packg.strings import hash_object
from packg.typext import PathType


class CachePlugin:
    """
    Helper to create a cache for a class.

    cache_name and cache_kwargs define a unique filename.
    cache_kwargs must be jsonable
    """

    def __init__(
        self,
        cache_name: str,
        cache_kwargs: Optional[dict] = None,
        cache_dir: Optional[PathType] = None,
        map_size=10 * 1024**3,  # 10 GB max size of database, increase if necessary
        verbose: bool = False,
    ):
        if cache_dir is None:
            cache_dir = get_cache_dir()
        if cache_kwargs is None:
            kwargs_str = "none"
        else:
            kwargs_str = hash_object(cache_kwargs)

        lmdb_file = Path(cache_dir) / "cache_plugin" / f"{cache_name}~{kwargs_str}.lmdb"
        os.makedirs(lmdb_file.parent, exist_ok=True)
        self.db: lmdb.Environment = lmdb.open(lmdb_file.as_posix(), map_size=map_size)
        self.lmdb_file = lmdb_file
        if verbose:
            print(f"Using cache file: {self.lmdb_file}")

    def get_values(self, list_of_keys):
        outputs = []
        with self.db.begin(write=False) as txn:
            for key in list_of_keys:
                bkey = msgpack.packb(key)
                bvalue = txn.get(bkey)
                if bvalue is None:
                    outputs.append(None)
                else:
                    value = msgpack.unpackb(bvalue)
                    outputs.append(value)
        return outputs

    def put_values(self, list_of_keys, list_of_values):
        assert len(list_of_keys) == len(
            list_of_values
        ), f"Got {len(list_of_keys)} keys and {len(list_of_values)} values."
        if len(list_of_keys) == 0:
            return
        with self.db.begin(write=True) as txn:
            for key, value in zip(list_of_keys, list_of_values):
                bkey = msgpack.packb(key)
                bvalue = msgpack.packb(value)
                txn.put(bkey, bvalue)



from .db_opener import DBOpener
from .sqlalchemist import (
    convert_sqlalchemy_orm_objects_to_dict,
    bulk_insert_mappings_ignore_dups_postgresql,
    convert_sqlalchemy_rows_to_dict,
    pg_insert,
    split_postgresql_db_uri,
    get_table_creation_sql,
    get_orm_class_by_table_name,
    get_orm_column_types,
    convert_data_types,
    bulk_insert_mappings_ignore_dups_singlethreaded,
    get_deletion_order,
    pd_read_sql_table,
)

__all__ = [
    "DBOpener",
    "convert_sqlalchemy_orm_objects_to_dict",
    "bulk_insert_mappings_ignore_dups_postgresql",
    "convert_sqlalchemy_rows_to_dict",
    "pg_insert",
    "split_postgresql_db_uri",
    "get_table_creation_sql",
    "get_orm_class_by_table_name",
    "get_orm_column_types",
    "convert_data_types",
    "bulk_insert_mappings_ignore_dups_singlethreaded",
    "get_deletion_order",
    "pd_read_sql_table",
]



from collections import defaultdict

import pandas as pd
import re
from copy import deepcopy
from io import StringIO
from sqlalchemy import ARRAY
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.engine.row import Row
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import (
    InstrumentedAttribute,
    ColumnProperty,
    Session,
)
from sqlalchemy.schema import CreateTable, Column
from types import GenericAlias
from typing import get_args, get_origin

from packg.log import logger


def get_orm_class_by_table_name(base, table_name):
    for cls in base.registry.mappers:
        if cls.class_.__tablename__ == table_name:
            return cls.class_
    raise ValueError(f"No ORM class found for table name {table_name}")


def convert_sqlalchemy_orm_objects_to_dict(result: list[object]):
    """
    Convert SQLAlchemy data objects to dictionaries. These are still database objects e.g.
    medpc.pacs.database_model.Studies and not Row.

    Here, Studies is a mapped class (likely an ORM class). This means that studies[0] is an instance
    of the Studies class, which corresponds to a row in the database. The object has attributes
    that correspond to the columns in the Studies table, and you can access them as object
    attributes, like studies[0].StudyDate.
    """
    output = []
    for i, r in enumerate(result):
        try:
            r_dict = r.__dict__
        except AttributeError as e:
            raise ValueError(
                f"Object {r} type {type(r)} in list at position {i} does not have a __dict__ "
                f"attribute so it is not an ORM object as expected. If the type is "
                f"sqlalchemy.engine.row.Row then the solution is to use the function "
                f"convert_sqlalchemy_rows_to_dict instead."
            ) from e
        r_dict_new = {k: v for k, v in r_dict.items() if not k.startswith("_")}
        output.append(r_dict_new)
    return output


def convert_sqlalchemy_rows_to_dict(result: list[Row]):
    """
    Convert SQLAlchemy Row objects to dictionaries. These are results which have been completely
    resolved and fetched, so there are no more dependencies to other tables.

    In this case, the table selected from is from SQLAlchemy's Table construct (not an ORM
    class). When you query it, the result is not an instance of a mapped class but a Row object.
    The Row object is a key-value mapping of column names to values, and you access data using
    dictionary-like syntax, like row["StudyDescription"].
    """
    # return [dict(row._mapping) for row in result]  # noqa
    output = []
    for i, r in enumerate(result):
        try:
            r_dict = dict(r._mapping)  # noqa
        except AttributeError as e:
            raise ValueError(
                f"Object {r} type {type(r)} in list at position {i} does not have a _mapping "
                f"attribute so it is not a Row object as expected. If the type is an sqlalchemy "
                f"ORM object like module.database_model.TableName then the solution is to use the "
                f"function convert_sqlalchemy_orm_objects_to_dict instead."
            ) from e
        r_dict_new = {k: v for k, v in r_dict.items() if not k.startswith("_")}
        output.append(r_dict_new)
    return output


def bulk_insert_mappings_ignore_dups_singlethreaded(
    session: Session, orm_table_class, unique_id_column_str, data: list[dict]
):
    """
    insert with e.g. sqlite and ignore duplicates by checking for the ids in the table. in
    multi-write this might crash if new duplicates are inserted between select and insert

    Args:
        session: sqlalchemy session
        orm_table_class: Studies
        unique_id_column_str: "StudyInstanceUID"
        data: list of dict with each dict representing a row to insert

    Returns:

    """
    # remove duplicates in the input data
    id2data = {}
    for d in data:
        did = d[unique_id_column_str]
        if did in id2data:
            logger.warning(
                f"Duplicate input id: {unique_id_column_str}={did} found. Ignoring the second "
                f"entry.\nExisting: {id2data[did]}\nDuplicate: {d}"
            )
            continue
        id2data[did] = d
    ids = list(id2data.keys())

    # select existing ids from db and skip all input data that already exists
    column = getattr(orm_table_class, unique_id_column_str)
    existing_ids_rows = session.query(column).filter(column.in_(ids)).all()
    existing_ids = set(
        [a[unique_id_column_str] for a in convert_sqlalchemy_rows_to_dict(existing_ids_rows)]
    )
    new_data = [d for did, d in id2data.items() if did not in existing_ids]
    if len(new_data) != len(data):
        logger.warning(
            f"Ignoring {len(data) - len(new_data)} duplicates out of {len(data)} "
            f"records for table {orm_table_class.__tablename__}"
        )

    # sqlite does not give enough information on errors. catch the error and reraise with more info
    errors_to_check = (OverflowError, IntegrityError)
    try:
        session.bulk_insert_mappings(orm_table_class, new_data)
    except errors_to_check as e:
        session.rollback()
        logger.error(f"Error inserting data: {type(e).__name__} {e}")
        logger.info(f"Inserting one by one to find the problematic entry...")
        for datapoint in new_data:
            try:
                session.bulk_insert_mappings(orm_table_class, [datapoint])
            except errors_to_check as e:
                logger.error(f"Error inserting single datapoint: {type(e).__name__} {e}")
                logger.error(f"Problematic entry: {datapoint}")
                raise e
        raise RuntimeError("The error above did not reappear when inserting data 1 by 1!") from e


def bulk_insert_mappings_ignore_dups_postgresql(
    session: Session,
    orm_table_class,
    index_elements: str | list[str],
    data: list[dict],
    count_inserts: bool = False,
    dedup_input: bool = True,
):
    """
    postgresql dialect has a special insert statement that can ignore duplicates.
    with count_inserts true

    Args:
        session: sqlalchemy session
        orm_table_class: Studies
        index_elements: name of the unique id column(s)
        data: list of dict with each dict representing a row to insert
        count_inserts: if True, count and log the number of ignored duplicates.
            this is ~10-20x slower than without counting and logging.
        dedup_input: if True, remove duplicates in the input data
    Returns:

    """
    n_inputs_before_dedup = len(data)
    if isinstance(index_elements, str):
        if dedup_input:
            id2data = {d[index_elements]: d for d in data}
            n_inputs_before_dedup = len(data)
            data = list(id2data.values())
        index_elements = [index_elements]
    else:
        index_elements = list(index_elements)
        if dedup_input:
            id2data = {tuple(d[i] for i in index_elements): d for d in data}
            n_inputs_before_dedup = len(data)
            data = list(id2data.values())

    stmt = pg_insert(orm_table_class).values(data)
    # count the inserts
    stmt = stmt.on_conflict_do_nothing(index_elements=index_elements)
    if not count_inserts:
        session.execute(stmt)
        return
    orm_field = getattr(orm_table_class, index_elements[0])
    stmt = stmt.returning(orm_field)
    result = session.execute(stmt)
    num_inserts = len(list(result))
    if num_inserts < len(data):
        i_in = len(data) - n_inputs_before_dedup
        i_db = n_inputs_before_dedup - num_inserts
        logger.warning(f"Input data: {len(data)}, ignored in input: {i_in}, already in db: {i_db}")


def get_orm_column_types(orm_table_class) -> dict[str, type]:
    """

    Args:
        orm_table_class:

    Returns:
        dict of column -> column type. the column type is either a standard python type like
        str, int, float or a type annotation like list[str], list[int], list[float]

    """
    field2type = {}
    supported_types = set((str, int, float, bool))
    for field, value in orm_table_class.__dict__.items():
        if field.startswith("_"):
            continue
        value: InstrumentedAttribute
        if not isinstance(value.property, ColumnProperty):
            # ignore relationships etc.
            continue
        proper: ColumnProperty = value.property
        columns = proper.columns
        assert len(columns) == 1, f"Expected one column for {field} but got {len(columns)}"
        column: Column = columns[0]
        if isinstance(column.type, ARRAY):
            array_content = column.type.item_type.python_type
            field2type[field] = list[array_content]
        elif column.type.python_type in supported_types:
            field2type[field] = column.type.python_type
        else:
            raise ValueError(f"Unsupported type {column.type} for field {field}")
    return field2type


def replace_0x00(x):
    # avoid postgresql error: ValueError: A string literal cannot contain NUL (0x00)
    if isinstance(x, str):
        return x.replace("\x00", " ")


def convert_data_types(orm_table_class, data: list[dict], str_replace_0x00: bool = False):
    """
    Convert data to the table structure. E.g. if input is int, and table field is str, convert
    to str. E.g.: studies_insert_list = convert_data_types(Studies, studies_insert_list)
    """
    field2type = get_orm_column_types(orm_table_class)
    new_data = []
    for d in data:
        dnew = deepcopy(d)
        for k in list(dnew.keys()):
            v = dnew[k]
            if v is None:
                continue
            target_type = field2type[k]
            if isinstance(target_type, GenericAlias) and get_origin(target_type) == list:
                # convert given a type like list[int]
                args = get_args(target_type)
                assert len(args) == 1, f"Expected 1 arg for typed list {target_type} but got {args}"
                inner_type = args[0]
                new_list = []
                for i in v:
                    if not isinstance(i, inner_type):
                        i = inner_type(i)
                    if str_replace_0x00:
                        i = replace_0x00(i)
                    new_list.append(i)
                v = new_list
            else:
                # convert a simple type like int, str, float
                if not isinstance(v, target_type):
                    v = target_type(v)
                if str_replace_0x00:
                    v = replace_0x00(v)
            dnew[k] = v
        new_data.append(dnew)
    return new_data


def _migrate_sqlite_to_postgresql_example():
    import sqlite3
    import pandas as pd

    # 1. connect to sqlite
    conn = sqlite3.connect("example.db")
    cursor = conn.cursor()

    # 2. list all existing tables
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = cursor.fetchall()
    print(tables)

    Example = None  # assuming this is the ORM class for the table
    session: Session = None  # assuming this session is connected to postgresql with sqlalchemy ORM
    chunk_size = 50000
    for table_name, table_orm in [
        ("example", Example),
    ]:
        print("*" * 70)
        print("*" * 70, table_name)
        print("*" * 70)
        total = 0
        # 3. select chunks
        for chunk in pd.read_sql_query(f"SELECT * FROM {table_name}", conn, chunksize=chunk_size):
            total += len(chunk)
            print(f"Chunk with {len(chunk)} rows loaded. total {total}")
            records: list[dict] = chunk.to_dict(orient="records")
            # 4. convert to target column types and insert
            records_conv = convert_data_types(table_orm, records, str_replace_0x00=True)
            session.bulk_insert_mappings(table_orm, records_conv)
            session.commit()

    # conn.close()


def get_table_creation_sql(db_model, engine):
    output = StringIO()
    for table in db_model.metadata.sorted_tables:
        sql = str(CreateTable(table).compile(engine))
        output.write(sql + ";\n\n")
    return output.getvalue()


def split_postgresql_db_uri(dburi: str) -> tuple[str, str, str, str, str]:
    """
    Split dburi into it's components, e.g. to give the components to psycopg2.connect()

    Args:
        dburi: postgresql+psycopg2://USER:PASS@SERVER:PORT/DBNAME

    Returns:
        tuple of (user, password, server, port, dbname)

    """
    re_db = re.compile(r"postgresql\+psycopg2://([^:]+):([^@]+)@([^:]+):(\d+)/(.+)")
    m = re_db.match(dburi)
    assert m is not None, f"DBURI {dburi} does not match the expected format"
    user, password, server, port, dbname = m.groups()
    return user, password, server, port, dbname


def get_deletion_order(Base) -> list[str]:
    """
    Get the order in which tables should be deleted to avoid foreign key constraint violations.

    Args:
        Base: declarative_base() object

    Returns:
        list of table names in the order they should be deleted

    """

    # Get all ORM classes (tables) from Base's metadata
    orm_classes = list(Base.metadata.sorted_tables)  # Sorted tables in the metadata

    # Build dependencies graph
    table_dependencies = defaultdict(set)
    for cls in orm_classes:
        # Inspect the columns for foreign key relationships
        for column in cls.columns:
            if column.foreign_keys:
                for fk in column.foreign_keys:
                    referenced_table = fk.column.table
                    # Add the dependent table to the list of dependencies for the referenced table
                    table_dependencies[referenced_table].add(cls)

    # Perform a topological sort on the table dependencies
    visited = set()
    deletion_order = []

    def visit(cls):
        if cls not in visited:
            visited.add(cls)
            # Recursively visit dependent tables
            for dependency in table_dependencies[cls]:
                visit(dependency)
            deletion_order.append(cls)

    # Start the visitation from all tables
    for cls in orm_classes:
        visit(cls)

    return [a.name for a in deletion_order]


def pd_read_sql_table(
    engine, base, table_name, sort_column: str | None = None, sort_ascending: bool = True
):
    table_orm = get_orm_class_by_table_name(base, table_name)
    df = pd.read_sql(f"SELECT * FROM {table_name}", engine)
    for column, dtype in get_orm_column_types(table_orm).items():
        if dtype == int:
            dtype = "Int64"
        if dtype == str:
            dtype = "string"
        if dtype == list or get_origin(dtype) == list:  # list or list[int] etc.
            # pandas does not support list of int
            dtype = "object"

        # print("map", column, "to", dtype)
        df[column] = df[column].astype(dtype)
    if sort_column is not None:
        df.sort_values(sort_column, inplace=True, ascending=sort_ascending)
        df.reset_index(drop=True, inplace=True)  # sorting only works together with index reset
        # assert df["index"].tolist() == list(range(df.shape[0])), "index not contiguous"
        # for i, row in df.iterrows():
        #     assert i == row["index"], f"{i=} {row['index']=}"
    return df



import os
import psutil
import pynvml  # pip install nvidia-ml-py
import torch
from loguru import logger
from typing import Optional, List

from packg import format_exception


def profile_ram():
    """

    Returns:
        RAM total (GB), RAM used (GB)
    """
    mem = psutil.virtual_memory()
    ram_total: float = mem.total / 1024**3
    ram_used: float = mem.used / 1024**3
    return ram_total, ram_used


def profile_ram_to_str():
    ram_total, ram_used = profile_ram()
    return f"RAM {ram_used:.3f}/{ram_total:.3f}GB"


def get_gpu_profiler():
    try:
        profiler = GPUProfilerNvml()
        return profiler
    except Exception as e:
        logger.error(
            f"Could not setup CUDA NVML GPU Profiler. Falling back to torch profiler. "
            f"Reason: {format_exception(e)}"
        )
        print(f"Log message")
    profiler = GPUProfilerTorch()
    return profiler


class GPUProfilerInterface:
    def profile_gpu(self, gpu_numbers: Optional[List[int]] = None):
        raise NotImplementedError

    def profile_to_str(self, gpu_numbers: Optional[List[int]] = None):
        """
        Use profile output to create a string

        Args:
            gpu_numbers: optional numbers of GPUs to profile

        Returns:
            profile string for output
        """
        ram_total, ram_used = profile_ram()
        profile_gpu_output = self.profile_gpu(gpu_numbers)
        if profile_gpu_output is None:
            return f"No GPUs detected!RAM {ram_used:.1f}/{ram_total:.1f}"

        names, mem_total, mem_used, load_gpu, load_gpu_mem, temp = profile_gpu_output
        # average / sum over all GPUs
        sum_mem_total: float = sum(mem_total)
        sum_mem_used: float = sum(mem_used)
        # gpu_mem_percent: float = gpu_mem_used / gpu_mem_total
        avg_load_gpu: float = sum(load_gpu) / max(1, len(load_gpu))
        avg_load_gpu_mem: float = sum(load_gpu_mem) / max(1, len(load_gpu_mem))
        avg_temp: float = sum(temp) / max(1, len(temp))

        # log the values
        multi_load, multi_load_mem, multi_temp, multi_mem = "", "", "", ""
        if len(load_gpu) > 1:
            multi_load = f' [{" ".join([f"{ld:.0%}" for ld in load_gpu])}]'
            multi_load_mem = f' [{" ".join([f"{ld:.0%}" for ld in load_gpu_mem])}]'
            multi_temp = f' [{" ".join([f"{ld:d}" for ld in temp])}]'
            multi_mem = f' [{" ".join([f"{mem:.1f}GB" for mem in mem_used])}]'

        out_str = " ".join(
            (
                # f"GPU {gpu_names_str}",
                f"Load {avg_load_gpu:.0%}{multi_load}",
                f"UMem {avg_load_gpu_mem:.0%}{multi_load_mem}",
                f"Mem {sum_mem_used:.1f}/{sum_mem_total:.1f}{multi_mem}",
                f"Temp {avg_temp:.0f}°C{multi_temp}",
                f"RAM {ram_used:.1f}/{ram_total:.1f}",
            )
        )

        return out_str

    def get_gpu_numbers(self):
        """
        Try to determine gpu numbers from env variables, fallback is to profile all gpus

        Returns:
            List with GPU numbers
        """
        gpu_count = self.gpu_count  # noqa
        gpu_numbers = list(range(gpu_count))
        device_env = os.getenv("CUDA_VISIBLE_DEVICES")
        if device_env is not None:
            try:
                gpu_numbers = [int(x.strip()) for x in device_env.split(",") if x.strip() != ""]
            except Exception:
                pass
            if len(gpu_numbers) == 0:
                gpu_numbers = list(range(gpu_count))
        return gpu_numbers

    def get_gpu_names(self, gpu_numbers: Optional[List[int]] = None):
        names, mem_total, mem_used, load_gpu, load_gpu_mem, temp = self.profile_gpu(
            gpu_numbers=gpu_numbers
        )
        return names

    def profile_single_gpu(self, gpu_number: int = 0):
        outputs = self.profile_gpu([gpu_number])
        return [x[0] for x in outputs]

    def check_gpus_have_errors(self, gpu_numbers: Optional[List[int]] = None):
        raise NotImplementedError


class GPUProfilerTorch(GPUProfilerInterface):
    def __init__(self):
        self.gpu_count = torch.cuda.device_count()
        if self.gpu_count == 0:
            self.gpu_count = 1
        self.has_nvml = True
        try:
            pynvml.nvmlInit()
        except Exception as e:
            logger.error(
                f"Torch profiler cannot use nvml. Utilization and temperature values will not be "
                f"displayed. Reason: {format_exception(e)}"
            )
            self.has_nvml = False

    def profile_gpu(self, gpu_numbers: Optional[List[int]] = None):
        """
        Profile GPU utilization

        Args:
            gpu_numbers: optional numbers of GPUs to profile

        Returns:
            Lists with entries for each GPU with content:
                name, total memory (GB), used memory (GB), gpu load (0-1), memory load (0-1), temperature (°C)

        """
        gpu_numbers = self.get_gpu_numbers() if gpu_numbers is None else gpu_numbers
        devices = [f"cuda:{n}" for n in gpu_numbers]
        try:
            props = [torch.cuda.get_device_properties(device) for device in devices]
        except RuntimeError as e:
            logger.error(format_exception(e))
            return None
        mem_total = [p.total_memory / 1024**3 for p in props]
        mem_used = [torch.cuda.max_memory_allocated(device) / 1024**3 for device in devices]
        names = [p.name for p in props]
        if self.has_nvml:
            temp = [torch.cuda.temperature(device) for device in devices]
            load_gpu = [torch.cuda.utilization(device) / 100 for device in devices]
            load_gpu_mem = [torch.cuda.memory_usage(device) for device in devices]
        else:
            temp = [-273] * len(devices)
            load_gpu = [-1] * len(devices)
            load_gpu_mem = [-1] * len(devices)
        # safeguard in case names comes back as bytes (can happen for nvml, not sure for torch)
        names_str = []
        for name in names:
            if isinstance(name, bytes):
                name = name.decode("utf-8")
            names_str.append(name)
        return names_str, mem_total, mem_used, load_gpu, load_gpu_mem, temp

    def check_gpus_have_errors(self, gpu_numbers: Optional[List[int]] = None):
        gpu_numbers = self.get_gpu_numbers() if gpu_numbers is None else gpu_numbers
        has_errors = False
        for i in gpu_numbers:
            try:
                torch.cuda.get_device_properties(f"cuda:{i}")
            except Exception as e:
                logger.error(f"GPU {i} has error: {format_exception(e)}")
                has_errors = True
        return has_errors


class GPUProfilerNvml(GPUProfilerInterface):
    """
    Get information about GPU and RAM usage as string.

    Setup:
        pip install nvidia-ml-py pynvml

    Notes:
        gpu_load:
            Percent of time over the past second in which any work has been executing on the GPU.
        gpu_memory_load:
            Percent of time over the past second in which any framebuffer memory has been read or stored.

    Usage:
        >>> gpu_profiler = GPUProfilerNvml()
        >>> print(f"Detected GPUs: {gpu_profiler.get_gpu_names()}")
        >>> print(gpu_profiler.profile_to_str())

    """

    def __init__(self):
        self.gpu_count = -1
        self.gpu_handles = []

        pynvml.nvmlInit()
        self.gpu_count = pynvml.nvmlDeviceGetCount()
        if self.gpu_count == 0:
            self.gpu_count = 1
        self.gpu_handles = []
        for i in range(self.gpu_count):
            try:
                handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            except Exception as e:
                print(f"WARNING: Could not get handle for GPU {i} due to {format_exception(e)}")
                handle = None
            self.gpu_handles.append(handle)

    def profile_gpu(self, gpu_numbers: Optional[List[int]] = None):
        """
        Profile GPU utilization

        Args:
            gpu_numbers: optional numbers of GPUs to profile

        Returns:
            Lists with entries for each GPU with content:
                name, total memory (GB), used memory (GB), gpu load (0-1), memory load (0-1), temperature (°C)

        """
        gpu_numbers = self.get_gpu_numbers() if gpu_numbers is None else gpu_numbers
        gpu_numbers = [i for i in gpu_numbers if self.gpu_handles[i] is not None]

        mem_objs = [pynvml.nvmlDeviceGetMemoryInfo(self.gpu_handles[i]) for i in gpu_numbers]
        mem_total = [mem_obj.total / 1024**3 for mem_obj in mem_objs]
        mem_used = [mem_obj.used / 1024**3 for mem_obj in mem_objs]
        names = [pynvml.nvmlDeviceGetName(self.gpu_handles[i]) for i in gpu_numbers]
        load_objs = [pynvml.nvmlDeviceGetUtilizationRates(self.gpu_handles[i]) for i in gpu_numbers]
        load_gpu = [load_obj.gpu / 100 for load_obj in load_objs]
        load_gpu_mem = [load_obj.memory / 100 for load_obj in load_objs]
        temp = [
            pynvml.nvmlDeviceGetTemperature(self.gpu_handles[i], pynvml.NVML_TEMPERATURE_GPU)
            for i in gpu_numbers
        ]
        names_str = []
        for name in names:
            # on some systems the names come back as bytes, make sure they are string
            if isinstance(name, bytes):
                name = name.decode("utf-8")
            names_str.append(name)
        return names_str, mem_total, mem_used, load_gpu, load_gpu_mem, temp

    def check_gpus_have_errors(self, gpu_numbers: Optional[List[int]] = None):
        """
        Some GPUs can silently not work, and then jax will silently switch to CPU.
        This function checks for these errors.

        Args:
            gpu_numbers: optional numbers of GPUs to profile

        Returns:
            profile string for output
        """
        gpu_numbers = self.get_gpu_numbers() if gpu_numbers is None else gpu_numbers

        has_errors = False
        for i in gpu_numbers:
            if self.gpu_handles[i] is None:
                has_errors = True
                logger.error(f"ERROR: GPU {i} handle could not be created at startup.")
                continue
            try:
                pynvml.nvmlDeviceGetClockInfo(self.gpu_handles[i], pynvml.NVML_CLOCK_GRAPHICS)
            except Exception as e:
                logger.error(f"GPU {i} has error: {format_exception(e)}")
                has_errors = True
        return has_errors



import os

import pandas as pd
import sqlite3
from attrs import define
from loguru import logger
from pathlib import Path

from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from typedparser import VerboseQuietArgs, add_argument, TypedParser


@define
class Args(VerboseQuietArgs):
    sqlite_file: Path = add_argument(positional=True, help="sqlite file")
    run_export: bool = add_argument(action="store_true", help="Run the export", shortcut="-e")


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    conn = sqlite3.connect(args.sqlite_file)

    tables = pd.read_sql_query("SELECT name FROM sqlite_master WHERE type='table';", conn)
    print("Tables in the database:")
    print(tables)

    # For each table, display the first few rows
    for table_name in tables["name"]:
        print(f"\nTable: {table_name}")
        df = pd.read_sql_query(f"SELECT * FROM {table_name} LIMIT 5;", conn)
        print(df)
        cursor = conn.execute(f"SELECT COUNT(*) FROM {table_name}")
        count = cursor.fetchone()[0]
        print(f"Count: {count}")

    if args.run_export:
        export_dir = Path(f"{args.sqlite_file}_export")
        if export_dir.is_dir():
            logger.warning(f"Export directory already exists: {export_dir}")
        os.makedirs(export_dir, exist_ok=True)
        for table_name in tables["name"]:
            print(f"\nTable: {table_name}")
            df = pd.read_sql_query(f"SELECT * FROM {table_name}", conn)
            print(df.shape)
            csv_file = export_dir / f"{table_name}.csv"
            df.to_csv(csv_file, sep=",", index=False)
            print(f"Exported to: {csv_file}")

    conn.close()


if __name__ == "__main__":
    main()



import torch
from typing import Optional

from visiontext.htmltools import NotebookHTMLPrinter
from visiontext.mathutils import torch_stable_softmax


def show_classification_logits_as_html(
    logits,
    classnames,
    temp: float = 1.0,
    k: int = 5,
    true_classname: Optional[str] = None,
    true_classid: Optional[int] = None,
    font_size=1.0,
):
    pr = NotebookHTMLPrinter()
    pr.open_table(font_size=font_size)
    # pr.add_table_row(*["#", "P", f"")
    for rank, _class_idx, prob, name in build_classification_logits_table(
        logits, classnames, temp=temp, k=k, true_classname=true_classname, true_classid=true_classid
    ):
        is_header = False
        if true_classname is not None and name == true_classname:
            is_header = True
        if true_classid is not None and _class_idx == true_classid:
            is_header = True
        pr.add_table_row(*[rank + 1, f"{prob:5.1%}", name], is_header=is_header)
    pr.close_table(output=True)


def build_classification_logits_table(
    logits,
    classnames,
    temp: float = 1.0,
    k: int = 5,
    true_classname: Optional[str] = None,
    true_classid: Optional[int] = None,
):
    """

    Args:
        logits: logits over the classes, array shape (n_classes,)
        classnames: list of classnames
        temp: temperature for softmax
        k: top k to return
        true_classname: classname of true class to add to the visualization
        true_classid: classid of true class to add to the visualization

    Returns:
        list length K of tuple: [(class_index, prob, classname)]
    """
    assert not (
        true_classname is not None and true_classid is not None
    ), "Cannot specify both true_classname and true_classid"
    logits_sm = torch_stable_softmax(logits, temp=temp)
    # values, indices = torch.topk(logits_sm, k, dim=0)
    values, indices = torch.topk(logits_sm, len(classnames), dim=0)
    logits_sm_np = logits_sm.cpu().numpy()
    indices = indices.cpu().numpy()
    output = []
    added_true_class = False
    for i in range(k):
        class_index = indices[i]
        class_name = classnames[class_index]
        output.append((i, class_index, logits_sm_np[class_index], class_name))
        if true_classname is not None and class_name == true_classname:
            added_true_class = True
        if true_classid is not None and class_index == true_classid:
            added_true_class = True
    if (true_classname is not None or true_classid is not None) and not added_true_class:
        for i in range(len(classnames)):
            class_index = indices[i]
            class_name = classnames[class_index]
            if (true_classname is not None and class_name == true_classname) or (
                true_classid is not None and class_index == true_classid
            ):
                output.append((i, class_index, logits_sm_np[class_index], class_name))
                break

    return output



from __future__ import annotations

from collections import defaultdict

import re
import torch


def compare_statedicts(
    sd1: dict[str, torch.Tensor],
    sd2: dict[str, torch.Tensor],
    verbose: bool = False,
    atol=1e-8,
    rtol=1e-5,
) -> bool:
    """
    Compare two state dicts.

    Comparison math is similar to
    https://pytorch.org/docs/stable/generated/torch.allclose.html
    """
    errors = []
    set_sd1 = set(sd1.keys())
    set_sd2 = set(sd2.keys())
    # mismatches = set_sd1 ^ set_sd2  # symmetric difference operator
    keys_in_1_not_2 = set_sd1 - set_sd2
    if len(keys_in_1_not_2) > 0:
        errors.append(f"Keys in sd1 but not sd2: {sorted(keys_in_1_not_2)}")
    keys_in_2_not_1 = set_sd2 - set_sd1
    if len(keys_in_2_not_1) > 0:
        errors.append(f"Keys in sd2 but not sd1: {sorted(keys_in_2_not_1)}")
    matches = set_sd1 & set_sd2  # intersection operator
    all_diffs = []
    for key in matches:
        t1 = sd1[key]
        t2 = sd2[key]
        if t1.shape != t2.shape:
            errors.append(f"Shape mismatch for key {key}: {t1.shape} != {t2.shape}")
            continue
        abs_diff = torch.abs(t1 - t2)
        all_diffs.append(abs_diff.reshape(-1))
        if abs_diff.max() > atol:
            errors.append(f"Max abs diff for key {key}: {abs_diff.max()} > {atol}")
        if abs_diff.max() > rtol * torch.abs(t2).max():
            errors.append(
                f"Max rel diff for key {key}: {abs_diff.max()} > {rtol * torch.abs(t2).max()}"
            )
    all_diffs = torch.cat(all_diffs)
    if len(errors) > 0:
        errors.append(f"Max abs diff: {all_diffs.max()} mean abs diff: {all_diffs.mean()}")
        if verbose:
            print("\n".join(errors))
        return False
    return True


def _group_by_requires_grad(param):
    if param.requires_grad:
        return "grad"
    return "no_grad"


def _get_params(parameters):
    # support inputs: model, model.parameters(), model.named_parameters()
    if hasattr(parameters, "values"):
        parameters = parameters.values()
    if hasattr(parameters, "parameters"):
        parameters = parameters.parameters()
    return parameters


def count_params_by_requires_grad(parameters) -> dict[str, int]:
    parameters = _get_params(parameters)
    groups = {"grad": 0, "no_grad": 0}
    for v in parameters:
        if v.requires_grad:
            groups["grad"] += v.numel()
        else:
            groups["no_grad"] += v.numel()
    return groups


def count_params(parameters) -> int:
    parameters = _get_params(parameters)
    total = 0
    for v in parameters:
        total += v.numel()
    return total


def group_params_and_shapes_for_display(param_names, param_shapes):
    """
    Compress parameter names by grouping those that differ only in their first numeric part.

    Args:
        param_names: List of parameter names
        param_shapes: List of parameter shapes

    Returns:
        Tuple of lists (new_compressed_param_names, param_shapes)
    """
    groups = defaultdict(list)
    ungrouped = []

    # For each parameter, find the first number in its name
    for name, shape in zip(param_names, param_shapes):
        m = re.search(r"(\d+)", name)
        if m:
            # Split param into prefix, the found number, and suffix
            start, end = m.span(1)
            prefix = name[:start]
            number_str = m.group(1)
            suffix = name[end:]
            block_number = int(number_str)
            # Use the shape, prefix, and suffix as the grouping key
            key = (shape, prefix, suffix)
            groups[key].append(block_number)
        else:
            # If no number is found, leave this parameter ungrouped
            ungrouped.append((name, shape))

    output = []
    # Process groups that have a numeric part
    for (shape, prefix, suffix), nums in groups.items():
        nums.sort()
        if len(nums) > 1:
            # Compress multiple block numbers into a range (min-max)
            compressed_number = f"[{nums[0]}-{nums[-1]}]"
        else:
            compressed_number = str(nums[0])
        output.append((f"{prefix}{compressed_number}{suffix}", shape))

    # Append any entries that didn't contain a number
    output.extend(ungrouped)
    return (list(a) for a in zip(*output))



from __future__ import annotations
from collections import defaultdict

import json
import os
import sqlite3
import time
from pathlib import Path
from timeit import default_timer
from typing import List, Tuple, Union, Optional

from visiontext.iotools.tar_indexer import index_tar
from visiontext.distutils import is_main_process, barrier_safe


class TarLookup:
    def __init__(
        self,
        base_path: Union[str, Path],
        tar_files_rel: List[Union[str, Path]],
        index_file: Union[str, Path],
        verbose: bool = True,
        sort_tar_files: bool = True,
        worker_id: int = 0,
        delete_index: bool = False,
    ):
        """
        Allow to lookup files inside tar archives via random access.

        Note: This class does not have update functionality. So if the tars change
        the index must be deleted.

        Args:
            base_path: Path where tar files are stored. The index will reference files relative
                to this path.
            tar_files_rel: List of tar files to index, relative to base_path OR absolute but inside
                base_path.
            index_file: Absolute path to sqlite index file.
            verbose: Print verbose output.
            sort_tar_files: Sort tar file input list before indexing.
            worker_id: Worker id to use for file pointers. When using multiprocessing dataloaders,
                each worker that uses this lookup needs its own file pointer to the tars.
                In torch use following code to find it:
                    from torch.utils.data import get_worker_info
                    worker = get_worker_info()
                    worker = worker.id if worker else None
            delete_index: Delete index file(s) and rebuild.

        """
        base_path = Path(base_path)
        index_file = Path(index_file)

        if verbose:
            print(
                f"Creating TarLookup with index {index_file} and base path {base_path} "
                f"with {len(tar_files_rel)} tar files."
            )
        if len(tar_files_rel) == 0:
            raise ValueError(f"No tar files to index for base path {base_path}.")

        filename_cache_file = Path(f"{index_file.as_posix()}.filenames.json")
        if delete_index:
            if verbose:
                print(f"Deleting {index_file} and {filename_cache_file} if they exist.")
            if index_file.is_file():
                index_file.unlink()
            if filename_cache_file.is_file():
                filename_cache_file.unlink()

        if not index_file.is_file():
            # index the tars
            if is_main_process():
                os.makedirs(base_path, exist_ok=True)
                os.makedirs(index_file.parent, exist_ok=True)
                if sort_tar_files:
                    tar_files_rel = sorted(tar_files_rel)
                for i, tar_file_rel in enumerate(tar_files_rel):
                    tar_file_rel = Path(tar_file_rel)
                    if tar_file_rel.is_absolute():
                        try:
                            tar_file_rel = tar_file_rel.relative_to(base_path)
                        except ValueError as e:
                            raise ValueError(
                                f"Tar file {tar_file_rel} must be relative to base_path {base_path}."
                            ) from e
                    tar_file_abs = base_path / tar_file_rel
                    if not tar_file_abs.is_file():
                        raise FileNotFoundError(f"Tar file {tar_file_abs} not found.")
                    if verbose:
                        print(
                            f"Indexing file {i}/{len(tar_files_rel)} dir {base_path} "
                            f"filename {tar_file_rel} to {index_file}"
                        )
                    # create sqlite index
                    index_tar(base_path, tar_file_rel, index_file, verbose=verbose)
            barrier_safe()

        # connect to database and load content
        try:
            connection = sqlite3.connect(str(index_file), check_same_thread=False)
        except sqlite3.OperationalError as e:
            print(f"Could not load {index_file}.")
            raise e
        cursor = connection.cursor()

        t1 = default_timer()

        # how many tarfiles are stored in the index?
        # fields in file_data are (file_id, file_name, mtime)
        # here file_id and file_name are referring to the tar files themselves
        # not worth caching since there will never be so many tar files
        cursor.execute("SELECT COUNT(file_id) FROM file_data")
        self.n_tar_files = int(cursor.fetchone()[0])
        if verbose:
            print(f"{default_timer() -t1:7.1f}s: Counted {self.n_tar_files} tar files")

        # select the filenames to build the index for this dataset
        # fields in offset_data are (file_id, file_name, offset, file_size)
        # here, file_id and file_name are referring to content files inside the tar.
        # with a 2GB index, 1200 tar, 30M files inside those, this select takes ? alot of time
        # therefore it is worth caching
        if not filename_cache_file.is_file():
            sql = "SELECT file_name FROM offset_data ORDER BY file_name"
            file_names = [data[0] for data in cursor.execute(sql).fetchall()]
            json.dump(file_names, filename_cache_file.open("w", encoding="utf-8"))
            if verbose:
                print(f"{default_timer() -t1:7.1f}s: Built index of {len(file_names)} filenames.")
        else:
            file_names = json.load(filename_cache_file.open(encoding="utf-8"))
            if verbose:
                print(
                    f"{default_timer() -t1:7.1f}s: Loaded index of {len(file_names)} filenames "
                    f"from {filename_cache_file}."
                )

        self.n_content_files = len(file_names)
        self.filepointers = {}

        self.file_names = file_names
        self.verbose = verbose
        self.base_path = base_path
        self.worker_id = worker_id
        self.index_file = index_file
        self.cursor = cursor
        self.connection = connection

    def get_fileinfo_from_filename(
        self, filename_in: str, n_tries: int = 3
    ) -> Tuple[str, str, int, int]:
        """
        Yield file information matching file_name from TAR archives.

        Args:
            filename_in: filename to search
            n_tries: in case of sqlite bugging out, try again a few times

        Returns:
            filename, tarfilename, offset (position inside the tar), file_size

        """
        fails = 0
        while True:
            try:
                self.cursor.execute(
                    """SELECT o.file_name, f.file_name, offset, file_size
                        FROM offset_data as o JOIN file_data as f ON o.file_id=f.file_id
                        WHERE o.file_name=?""",
                    (filename_in,),
                )
                break
            except Exception as e:
                print(
                    f"WARNING: {e} for file {self.index_file} "
                    f"looking for {filename_in} in index."
                )
                fails += 1
                time.sleep(1)
                if fails >= n_tries:
                    raise e
        # get information needed to load the file
        fetch = self.cursor.fetchone()
        if fetch is None:
            raise KeyError(
                f"File {filename_in} not found in index {self.index_file}. "
                f"Try deleting the index to rebuild it automatically."
            )
        filename, tarfilename, offset, file_size = fetch
        return filename, tarfilename, offset, file_size

    def get_content_from_filename(
        self, filename_in: str, n_tries: int = 3
    ) -> Tuple[str, str, bytes]:
        """
        Yield files matching file_name from TAR archives.

        Args:
            filename_in: filename to search
            n_tries: in case of sqlite bugging out, try again a few times

        Returns:
            filename, tarfilename, content (bytes)
        """
        filename, tarfilename, offset, file_size = self.get_fileinfo_from_filename(
            filename_in, n_tries
        )

        tarfn_full = self.base_path / tarfilename
        # create one filepointer for each combination of worker and tarfile
        worker = self.worker_id
        key = (worker, tarfilename)
        if key not in self.filepointers:
            # open tarfile
            self.filepointers[key] = open(tarfn_full, "rb")

        tarf = self.filepointers[key]
        # seek
        tarf.seek(offset)
        # return content
        content = tarf.read(file_size)
        return filename, tarfilename, content

    def get_content_from_index(self, idx: int) -> Tuple[str, str, bytes]:
        """
        Load content from index (number of a single file).

        Args:
            idx:

        Returns:
            filename, tarfilename, content (bytes)

        """
        file_name = self.file_names[idx]
        return self.get_content_from_filename(file_name)

    def get_files_per_shard(
        self, endswith: Optional[str] = None, sorter: callable = sorted
    ) -> dict[str, List[str]]:
        """
        Group files by shard name.

        Args:
            endswith: to count only e.g. .jpg files
            sorter: the default sorted is fine for sorting tars named 00000.tar, 00001.tar, ...
                however if tar filenames do not have leading zeros, set it to natsort.natsorted

        Returns:
            dict of shardname to list of filenames
        """
        self.cursor.execute(
            "SELECT o.file_name, f.file_name, offset, file_size FROM offset_data as o JOIN file_data as f ON o.file_id=f.file_id"
        )
        fetches = self.cursor.fetchall()
        shard2fi = defaultdict(list)
        for filename, shardname, _, _ in fetches:
            if endswith is not None:
                if not filename.endswith(endswith):
                    continue
            shard2fi[shardname].append(filename)

        return dict(sorter(shard2fi.items(), key=lambda x: x[0]))

    def __len__(self):
        return self.n_content_files



from pathlib import Path
from typing import Optional

from attrs import define
from loguru import logger

from packg.iotools import dump_json
from packg.iotools.pathspec_matcher import PathSpecArgs
from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from typedparser import VerboseQuietArgs, add_argument, TypedParser
from visiontext.iotools.tar_lookup import TarLookup


@define
class Args(PathSpecArgs, VerboseQuietArgs):
    base_dir: Optional[Path] = add_argument(
        shortcut="-b", type=str, help="Source base dir", default=None
    )
    tar_files_glob: str = add_argument(
        shortcut="-t", type=str, help="Glob pattern for tar files", default="**/*.tar"
    )
    index_file: Optional[Path] = add_argument(type=str, help="Index file", default=None)


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")
    tar_files = list(args.base_dir.glob(args.tar_files_glob))
    logger.info(f"Found {len(tar_files)} tar files in {args.base_dir}.")
    if args.index_file is None:
        args.index_file = args.base_dir / "index.sqlite"
    logger.info(f"Creating tarlookup and indexing to {args.index_file}")
    tar_lookup = TarLookup(
        args.base_dir,
        tar_files,
        args.index_file,
        verbose=True,
        sort_tar_files=True,
        worker_id=0,
        delete_index=False,
    )
    logger.info(f"Got {len(tar_lookup)} files inside the tars.")
    plaintext_filenames_file = Path(f"{args.index_file.as_posix()}.json")
    dump_json(
        {"filenames": tar_lookup.file_names},
        plaintext_filenames_file,
        indent=2,
        custom_format=False,
    )


if __name__ == "__main__":
    main()



from PIL import Image, ImageDraw, ImageFont
from matplotlib import colors as mpl_colors
from typing import Dict, List, Tuple, Optional, Union

from packg.maths import clip_rectangle_coords
from visiontext.font import get_font_path_dejavusans
from visiontext.colormaps import (
    create_colormap_for_dark_background,
    create_colorgetter_from_colormap,
)


def create_bbox_images(
    bx: int,
    by: int,
    bw: int,
    bh: int,
    w: int,
    h: int,
    box_color=(0, 0, 255, 255),
    bbox_width: int = 5,
    mask_opacity: int = 64,
) -> Dict[str, Image.Image]:
    """
    Create box overlay images, originally used in the OVAD visualizer e.g.:
    https://lmb.informatik.uni-freiburg.de/resources/datasets/ovad/?imgnum=14#top

    This function creates transparent PNGs that can be overlayed on top of an image to show or
    highlight a bounding box.
    """
    # create one image with transparent background and one with half transparent
    img = Image.new("RGBA", (w, h), (0, 0, 0, 0))
    draw = ImageDraw.Draw(img)
    img2 = Image.new("RGBA", (w, h), (0, 0, 0, mask_opacity))
    draw2 = ImageDraw.Draw(img2)

    # boxes should be approximately the same width after scaling the image
    # so we have to scale the box thickness now
    target_w = 800
    rel_w = w / target_w  # if small image, create smaller boxes
    rel_bbox_width = round(bbox_width * rel_w)

    # the inside of the thick drawn box will be the true bbox
    # this means that the box can clip outside the image and become almost invisible
    # to avoid that, we move the edges of the box outwards and then clip them
    x1 = bx - rel_bbox_width
    y1 = by - rel_bbox_width
    x2 = bx + bw + rel_bbox_width
    y2 = by + bh + rel_bbox_width
    (x1, y1, x2, y2) = clip_rectangle_coords((x1, y1, x2, y2), w, h)

    for bbox_step in range(rel_bbox_width):
        rectangle_coords = (
            x1 + bbox_step,
            y1 + bbox_step,
            x2 - bbox_step,
            y2 - bbox_step,
        )
        # a 2nd clip should now not be necessary anymore and we assert instead
        assert (
            rectangle_coords[0] >= 0
            and rectangle_coords[1] >= 0
            and rectangle_coords[2] <= w
            and rectangle_coords[3] <= h
        ), f"{rectangle_coords} outside {w}x{h}"

        draw.rectangle(rectangle_coords, fill=(0, 0, 0, 0), outline=box_color)
        draw2.rectangle(rectangle_coords, fill=(0, 0, 0, 0), outline=box_color)

    return {"box": img, "box-mask": img2}


def convert_bbox_abs_to_rel(bx, by, bw, bh, image_w, image_h):
    """
    Convert bounding box coordinates to relative coordinates.

    Args:
        bx: x coordinate of the top-left corner of the bounding box
        by: y coordinate of the top-left corner of the bounding box
        bw: width of the bounding box
        bh: height of the bounding box
        image_w: width of the image
        image_h: height of the image

    Returns:
        Tuple of relative coordinates (x, y, w, h)

    """
    x = bx / image_w
    y = by / image_h
    w = bw / image_w
    h = bh / image_h
    return x, y, w, h


def convert_bbox_rel_to_abs(x, y, w, h, image_w, image_h):
    """
    Convert bounding box coordinates to absolute coordinates.

    Args:
        x: x coordinate of the top-left corner of the bounding box
        y: y coordinate of the top-left corner of the bounding box
        w: width of the bounding box
        h: height of the bounding box
        image_w: width of the image
        image_h: height of the image

    Returns:
        Tuple of absolute coordinates (bx, by, bw, bh)

    """
    bx = x * image_w
    by = y * image_h
    bw = w * image_w
    bh = h * image_h
    return bx, by, bw, bh


def get_bbox_bounds(
    box_x: float,
    box_y: float,
    box_w: float,
    box_h: float,
    image_w: int,
    image_h: int,
    min_w=32,
    min_h=32,
    create_squares=False,
):
    """
    Get the bounding box boundaries for cropping the image.
    Increase the box size if it is too small.

    Args:
        box_x: absolute box coordinates for the given image
        box_y:
        box_w:
        box_h:
        image_w: image size
        image_h:
        min_w: minimum box size
        min_h:
        create_squares: if True, try to create a square bounding box instead of a rectangle

    Returns:
        Cropping parameters tuple (x1, y1, x2, y2)

    """
    # Calculate the center coordinates of the bounding box
    center_x = box_x + (box_w / 2)
    center_y = box_y + (box_h / 2)

    # Calculate the new width and height of the bounding box
    new_w = max(box_w, min_w)
    new_h = max(box_h, min_h)

    if create_squares:
        new_w = max(new_w, new_h)
        new_h = max(new_w, new_h)
        pass

    # Calculate the new x and y coordinates of the top-left corner of the bounding box
    new_x = center_x - (new_w / 2)
    new_y = center_y - (new_h / 2)

    # Adjust the new x coordinate if the box is too close to the left border
    if new_x < 0:
        new_x = 0
        new_x2 = min(new_x + new_w, image_w)
    else:
        new_x2 = min(new_x + new_w, image_w)
        if new_x2 > image_w:
            new_x2 = image_w
            new_x = max(new_x2 - new_w, 0)

    # Adjust the new y coordinate if the box is too close to the top border
    if new_y < 0:
        new_y = 0
        new_y2 = min(new_y + new_h, image_h)
    else:
        new_y2 = min(new_y + new_h, image_h)
        if new_y2 > image_h:
            new_y2 = image_h
            new_y = max(new_y2 - new_h, 0)

    # Again make sure all bounds are respected
    new_x = max(new_x, 0)
    new_y = max(new_y, 0)
    new_x2 = min(new_x2, image_w)
    new_y2 = min(new_y2, image_h)

    # Return the cropping boundaries as integers
    return int(new_x), int(new_y), int(new_x2), int(new_y2)


def draw_bounding_boxes_pil(
    image: Image.Image,
    boxes: List[Tuple[float, float, float, float]],  # x, y, w, h
    labels: Tuple[List[str], str, None] = "numbers",
    colors: Optional[List[Tuple[int, int, int]]] = None,
    width: int = 3,
    font_size: int = 14,
    cmap: Optional[Union[str, mpl_colors.Colormap]] = None,
):
    draw = ImageDraw.Draw(image)
    font = ImageFont.truetype(get_font_path_dejavusans(), font_size)

    if colors is None:
        if cmap is None:
            cmap = create_colormap_for_dark_background(add_brightness=0.5)
        # automatically generate a bunch of colors from the colormap
        getter = create_colorgetter_from_colormap(cmap)
        colors = [tuple(getter(i / len(boxes))) for i in range(len(boxes))]

    if isinstance(labels, str):
        if labels == "numbers":
            labels = [str(i) for i in range(len(boxes))]
        else:
            raise ValueError(f"Unknown label type: {labels}")

    for i, (box, color) in enumerate(zip(boxes, colors)):
        x, y, w, h = box
        x1, y1, x2, y2 = get_bbox_bounds(x, y, w, h, image.width, image.height, min_w=0, min_h=0)
        draw.rectangle([(x1, y1), (x2, y2)], outline=color, width=width)

        text_args = dict(font=font, stroke_width=2)
        if labels is not None:
            label = labels[i]

            # # removed in pillow 10...
            # text_width, text_height = draw.textsize(label, **text_args)

            # new version with textbbox
            tx0, ty0, tx1, ty1 = draw.textbbox((0, 0), label, **text_args)
            text_width = tx1 - tx0
            text_height = ty1 - ty0

            text_x = min(x1 + width + 2, image.width - text_width)
            text_y = min(y1 + width + 2, image.height - text_height)
            draw.text(
                (text_x, text_y),
                label,
                fill=color,
                anchor="lt",
                stroke_fill="black",
                **text_args,
            )

    return image



from pathlib import Path

from loguru import logger
from pyinstrument import Profiler, renderers

from packg.dtime import get_timestamp_for_filename

current_profiler: Profiler = None


def start_pyinstrument_profiler():
    global current_profiler
    if current_profiler is not None:
        logger.warning("Profiler already running!")
        return
    current_profiler = Profiler()
    current_profiler.start()


def stop_pyinstrument_profiler(
    open_in_browser=True, output_text=True, print_fn=print, unicode=True, color=True
) -> str:
    if current_profiler is None:
        logger.error("No profiler running!")
        return ""
    current_profiler.stop()
    text = current_profiler.output_text(unicode=unicode, color=color)
    if output_text:
        print_fn(text)
    if open_in_browser:
        # in ubuntu, firefox can't access /tmp - hack into pyinstrument and save to home
        # current_profiler.open_in_browser()  # <- doesn't work because it saves to /tmp/XXX.html
        tf = (Path.home() / f"py_profile_{get_timestamp_for_filename()}.html").as_posix()
        session = current_profiler._get_last_session_or_fail()  # noqa
        return renderers.HTMLRenderer(timeline=False).open_in_browser(session, output_filename=tf)
    return text



"""
Compute spectrogram of an audio file.
"""

from attrs import define
from loguru import logger
from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from pathlib import Path
from typedparser import VerboseQuietArgs, add_argument, TypedParser
from visiontext.audiotools import load_audio_file, AudioConvertC, check_audio_quality


@define
class Args(VerboseQuietArgs):
    input_file: Path = add_argument("input_file", type=str, help="Audio file to check")


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    frequency, data = load_audio_file(args.input_file, verbose=True, convert=AudioConvertC.MONO)

    print(f"quality: {check_audio_quality(frequency, data)}")

    # plot the spectogram
    # plt.specgram(data, Fs=frequency, NFFT=128, noverlap=0)  # plot
    # plt.show()


if __name__ == "__main__":
    main()



import lmdb
import msgpack
import os
from pathlib import Path
from typing import Optional

from packg.paths import get_cache_dir
from packg.strings import hash_object
from packg.typext import PathType


class LmdbAutoGrowEnvironment(lmdb.Environment):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        try:
            super().__setitem__(key, value)
        except lmdb.MapFullError:
            self.set_mapsize(self.info()["map_size"] * 2)
            super().__setitem__(key, value)



"""
simple tar index, with one json index for each tar file, and only one tar file read at a time.

compared to the sqlite based tar indexer this is simpler, but probably less efficient
for large datasets because you always need to read the full index before searching,
instead of relying on dataset selects.
"""

from __future__ import annotations

import tarfile
from natsort import natsorted
from pathlib import Path
from typing import Optional

from packg import format_exception
from packg.iotools import load_json, dump_json
from packg.log import logger
from packg.typext import PathType


class SingleTarLookup:
    def __init__(
        self,
        tar_file: PathType,
        index_file: Optional[PathType] = None,
        force_rebuild_index: bool = False,
        worker_id: int = 0,
        check_stat: bool = True,
    ):
        self.tar_file = Path(tar_file)
        self.index = get_tar_index(tar_file, index_file, force_rebuild_index, check_stat=check_stat)
        self.filenames = list(self.index["files"].keys())
        if len(self.filenames) == 0:
            raise ValueError(f"No files in tar: {tar_file}")
        self.filepointers = {}
        self.worker_id = worker_id

    def __repr__(self):
        type_name = type(self).__name__
        return (
            f"{type_name}(tar_file={self.tar_file}, "
            f"worker_id={self.worker_id}) with {len(self.filenames)} files, example first "
            f"file: {self.filenames[0]}"
        )

    def get_filenames(self):
        return self.filenames

    def get_file_info(self, filename_in: str) -> tuple[int, int, float]:
        if filename_in not in self.index["files"]:
            raise FileNotFoundError(f"{filename_in} not found in {self.tar_file}")
        return self.index["files"][filename_in]

    def get_file_content(self, filename_in: str) -> bytes:
        offset, size, mtime = self.get_file_info(filename_in)
        if self.worker_id not in self.filepointers:
            self.filepointers[self.worker_id] = self.tar_file.open("rb")
        tarf = self.filepointers[self.worker_id]
        tarf.seek(offset)
        content = tarf.read(size)
        return content

    def has_file(self, filename_in: str) -> bool:
        return filename_in in self.index["files"]

    def close(self):
        if self.worker_id in self.filepointers:
            self.filepointers[self.worker_id].close()
            del self.filepointers[self.worker_id]


def get_index_file_for_tar_file(tar_file: PathType) -> Path:
    tar_file = Path(tar_file)
    return Path(f"{tar_file.as_posix()}.json")


def get_tar_index(
    tar_file: PathType,
    index_file: Optional[PathType] = None,
    force: bool = False,
    check_stat: bool = True,
) -> dict[str, any]:
    """

    Args:
        tar_file: tar file to index
        index_file: where to save the index, default f"{tar_file}.json"
        force: always overwrite the index
        check_stat: verify tar file size and mtime in the index (default True is slower)

    Returns:
        dictionary of
            size: tar file size
            mtime: tar file modification time
            files: dictionary of file name to (offset, size) in the tar

    """
    tar_file = Path(tar_file)
    tar_size, tar_mtime = None, None
    if check_stat:
        assert tar_file.is_file(), f"File not found: {tar_file}"
        tar_file_stat = tar_file.stat()
        tar_size, tar_mtime = tar_file_stat.st_size, tar_file_stat.st_mtime

    if index_file is None:
        index_file = get_index_file_for_tar_file(tar_file)
    else:
        index_file = Path(index_file)

    if not force and index_file.is_file():
        existing_index = load_json(index_file)
        size, mtime = existing_index["size"], existing_index["mtime"]
        if not check_stat or (size == tar_size and mtime == tar_mtime):
            # index already exists and is up to date
            return existing_index

    # index must be rebuilt
    try:
        fileindex = build_tar_fileindex(tar_file)
    except tarfile.ReadError as e:
        logger.error(f"Failed to read tar file {tar_file}: {format_exception(e)}")
        raise
    index = {"size": tar_size, "mtime": tar_mtime, "files": fileindex}
    dump_json(index, index_file, verbose=False)
    return index


def build_tar_fileindex(tar_file: PathType) -> dict[str, tuple[int, int, float]]:
    tar = tarfile.open(tar_file)
    fileindex = {}
    for tarinfo in tar:
        if not tarinfo.isfile():
            continue
        fileindex[tarinfo.name] = [tarinfo.offset_data, tarinfo.size, tarinfo.mtime]
    return fileindex


def search_and_compress_files_to_tar(
    tar_file: PathType,
    base_dir: PathType,
    patterns: tuple[str] = ("**/*",),
    verify: bool = True,
    delete_files: bool = False,
    sort_fn: Optional[callable] = natsorted,
):
    rel_files = []
    for pattern in patterns:
        rel_files.extend([file.relative_to(base_dir).as_posix() for file in base_dir.glob(pattern)])
    if len(rel_files) == 0:
        raise FileNotFoundError(f"No files found to compress for\n{base_dir=}\n{patterns=}")
    if sort_fn is not None:
        rel_files = sort_fn(rel_files)
    logger.info(f"Found {len(rel_files)} files in {base_dir}, compressing to {tar_file}")
    compress_files_to_tar(tar_file, base_dir, rel_files, verify=verify, delete_files=delete_files)


def compress_files_to_tar(
    tar_file: PathType,
    base_dir: PathType,
    rel_files: list[PathType],
    delete_files: bool = False,
    verify: bool = True,
):
    if tar_file.is_file():
        if verify:
            assert verify_tar(tar_file, base_dir, rel_files), f"Verification failed for {tar_file}"
            return
        return
    # tar does not exist, create the tar
    if len(rel_files) == 0:
        raise FileNotFoundError(f"No files given to compress.")
    with tarfile.open(tar_file.as_posix(), "w") as tar:
        for file_rel in rel_files:
            file_full = base_dir / file_rel
            info = tar.gettarinfo(file_full.as_posix())
            # write dummy user info
            info.uid = 10999
            info.gid = 10999
            info.uname = "bigdata"
            info.gname = "bigdata"
            # info.mtime = file_fullstat().mtime
            # info.size = file_full.stat().st_size
            info.mode = 0o666
            # save relative paths
            info.name = file_rel
            tar.addfile(info, file_full.open("rb"))
    if verify:
        assert verify_tar(tar_file, base_dir, rel_files), f"Verification failed for {tar_file}"
    if delete_files:
        logger.warning(f"Deleting {len(rel_files)} files in {base_dir} after compression.")
        for file_rel in rel_files:
            (base_dir / file_rel).unlink()


def verify_tar(tar_file: PathType, base_dir: PathType, rel_files: list[PathType]) -> bool:
    # check tar contains all files from above, and only those
    file_missing = {f: True for f in rel_files}
    try:
        with tarfile.open(tar_file.as_posix(), "r") as tar:
            for tarinfo in tar:
                assert tarinfo.isfile()
                assert file_missing[tarinfo.name], f"Duplicate file: {tarinfo.name}"
                file_missing[tarinfo.name] = False
    except tarfile.ReadError as e:
        logger.error(f"Failed to read tar file {tar_file}: {format_exception(e)}")
        return False

    missing_files = [f for f, missing in file_missing.items() if missing]
    assert len(missing_files) == 0, f"Missing files: {missing_files}"

    # check all files have correct content
    try:
        lookup = SingleTarLookup(tar_file)
        assert natsorted(lookup.get_filenames()) == rel_files
        for file_rel in rel_files:
            content = lookup.get_file_content(file_rel)
            gt_content = (base_dir / file_rel).read_bytes()
            assert content == gt_content
            offset, size, mtime = lookup.get_file_info(file_rel)
            gt_stat = (base_dir / file_rel).stat()
            assert size == gt_stat.st_size
            assert mtime == gt_stat.st_mtime
    except AssertionError as e:
        logger.error(f"Verification failed for tar file {tar_file}: {format_exception(e)}")
        return False
    return True



import os

from pathlib import Path
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import DeclarativeBase, Session, sessionmaker

from packg.log import logger


class DBOpener:
    def __init__(
        self,
        db_model,
        db_url: str = None,
        sqlite_db_file: str | Path | None = None,
        engine_kwargs: dict | None = None,
    ):
        """
        Initialize the database manager either with an SQLite file, or any other SQL database URL.

        Args:
            db_model: DeclarativeBase for sqlalchemy ORM
            db_url: URL for the database e.g. 'postgresql+psycopg2://user:pw@server:5432/dbname'
            sqlite_db_file: Path to the SQLite database file instead of a URL
        """
        if db_url is not None and sqlite_db_file is not None:
            raise ValueError(f"Can only provide one of {db_url=} or {sqlite_db_file=}, got both.")
        if db_url is not None:
            self.db_url: str = db_url
            self.sqlite_db_file: Path | None = None
        elif sqlite_db_file is not None:
            self.db_url: str = (f"sqlite:///{self.sqlite_db_file.as_posix()}",)
            self.sqlite_db_file: Path = Path(sqlite_db_file)
        self.db_model: DeclarativeBase = db_model
        self.engine_kwargs: dict | None = {} if engine_kwargs is None else engine_kwargs
        self.engine: Engine = None
        self.session: Session = None

    def connect(self, create_all: bool = True):
        """Connect to the SQLite database."""
        if self.sqlite_db_file is not None and not self.sqlite_db_file.is_file():
            logger.warning(f"Database file not found, recreating: {self.sqlite_db_file}")
            os.makedirs(self.sqlite_db_file.parent, exist_ok=True)
        logger.info(f"Connecting to database: {self.db_url}")
        self.engine: Engine = create_engine(
            self.db_url,
            # connect_args={"check_same_thread": False}
            **self.engine_kwargs,
        )
        self.session: Session = sessionmaker(bind=self.engine)()
        if create_all:
            self.db_model.metadata.create_all(self.engine, checkfirst=True)

        # self.connection = sqlite3.connect(self.db_file)
        # self.connection.row_factory = sqlite3.Row  # This allows us to get rows as dictionaries
        # self.cursor = self.connection.cursor()

    def close(self):
        """Close the database connection."""
        if self.session is not None:
            self.session.close()



from typing import Any

import os
import torch

from packg import format_exception
from packg.typext import PathType


def torch_save_safely(obj: Any, f: PathType, **kwargs):
    try:
        torch.save(obj, f, **kwargs)
    except KeyboardInterrupt as e:
        print(f"Deleting {f} due to KeyboardInterrupt")
        try:
            os.remove(f)
        except Exception as e2:
            print(f"Exception while deleting {f}: {format_exception(e2)}")
            pass
        raise e



from __future__ import annotations

import base64
import io
from IPython.display import display, HTML
from PIL import Image
from io import BytesIO
from matplotlib import pyplot as plt
from typing import Optional

from visiontext.images import scale_image_to_width, scale_image_to_height


class NotebookHTMLPrinter:
    """
    Class for printing things as HTML (Useful for jupyter notebooks)
    Uses an internal buffer to aggregate a bunch of HTML and then creates a single HTML output
    This is better than calling display(HTML(...))) separately for each line because it allows
    for things to stay inside one output cell.

    todo add option to add style to invdividual tr and td

    Usage:
        pr = NotebookHTMLPrinter()
        pr.print("<b>Hello World</b>", output=True)  # output=True to instantly display the HTML

        pr.open_table()
        pr.add_table_row("col1", "col2", "col3", is_header=True)
        pr.add_table_row("row1_col1", "row1_col2", "row1_col3")
        pr.close_table()  # here output=True is the default, so this will output the table

    """

    def __init__(self):
        self.reset()
        self.grid_max_width = 300

    def reset(self):
        self.buffer = []

    def print(self, *text, output=False, sep=" ", end="<br/>"):
        text = [str(t) for t in text]
        self.buffer.append(f"{sep.join(text)}{end}")
        if output:
            self.output()

    def get_html(self):
        html_txt = "".join(self.buffer)
        self.buffer = []
        return html_txt

    def output(self):
        display(HTML(self.get_html()))

    def open_table(self, font_size=1.5, other_style=""):
        """Start creating a HTML table e.g. for data output"""
        self.print(f"<table style='font-size:{font_size:.0%};{other_style}'>", end="")

    def add_table_row(self, *args, is_header=False, tr_style="", td_style=""):
        if tr_style is None or tr_style == "":
            tr_str = ""
        else:
            tr_str = f" style='{tr_style}'"
        if td_style is None or td_style == "":
            td_str = ""
        else:
            td_str = f" style='{td_style}'"

        tag = "th" if is_header else "td"
        self.print(f"<tr{tr_str}>", end="")
        for arg in args:
            self.print(f"<{tag}{td_str}>{arg}</{tag}>", end="")
        self.print("</tr>", end="")

    def close_table(self, output=True):
        self.print(f"</table>", output=output, end="")

    def open_grid(self, max_width=300, box_css="display: flex;", col_css=""):
        """Start creating a flexbox grid to output several HTML contents side by side"""
        self.grid_max_width = max_width
        self.print(f"<div style='{box_css}'>", end="")
        self._open_grid_column(col_css=col_css)

    def _open_grid_column(self, col_css=""):
        width_css = ""
        if self.grid_max_width is not None:
            width_css = f"max-width:{self.grid_max_width}px;"
        self.print(f"<div style='margin-right:10px;{width_css} {col_css}'>", end="")

    def _close_grid_column(self):
        self.print(f"</div>", end="")

    def next_grid_column(self, col_css=""):
        self._close_grid_column()
        self._open_grid_column(col_css=col_css)

    def close_grid(self, output=True):
        self._close_grid_column()
        self.print(f"</div>", output=output, end="")


def display_html_table(
    lines, header_row_indices=(0,), scroll_height=None, font_size=1.0, table_other_style=""
):
    """

    Args:
        lines: table data in format [[row1_col1, ..., row1_colN], ..., [rowM_col1, ..., rowM_colN]]
        header_row_indices: which rows should be considered headers and printed bold
        scroll_height: maximum height of the table in pixels, if None, infinite
        font_size: multiplier for font size of the table
        table_other_style: other style attributes for the table
    """
    header_row_indices = set() if header_row_indices is None else set(header_row_indices)
    p1 = NotebookHTMLPrinter()
    p1.open_table(font_size=font_size, other_style=table_other_style)
    for i, l in enumerate(lines):
        p1.add_table_row(*l, is_header=i in header_row_indices)
    p1.close_table(output=False)
    html_content = p1.get_html()
    display_scrollable(html_content, scroll_height=scroll_height)


def display_html_table_from_dict(table_dict, **kwargs):
    """
    Args:
        table_dict: dict of format
            {header1: [row1_col1, ..., row1_colN], ..., headerM: [rowM_col1, ..., rowM_colN]}
        **kwargs: passed to display_html_table
    """
    lines = [[k] + v for k, v in table_dict.items()]
    display_html_table(lines, **kwargs)


def display_scrollable(html_content, scroll_height=200):
    if scroll_height is not None and scroll_height > 0:
        # Puts the scrollbar next to the content
        css_style = "height: 200px; overflow: auto; width: fit-content; resize: both;"
        display(HTML(f"<div style='{css_style}'>{html_content}</div>"))
    else:
        display(HTML(html_content))


def get_colored_html_text(color_tuple, *text_args, sep=" ", end="\n"):
    assert len(color_tuple) in [1, 3, 4] and [
        0 <= ca <= 255 for ca in color_tuple
    ], f"Invalid rgb tuple: {color_tuple}"
    if len(color_tuple) == 1:
        color_tuple = color_tuple * 3
    hex_str = "".join([f"{ca:02x}" for ca in color_tuple])
    args_joined = sep.join(text_args)
    text = f"<span style='color:#{hex_str};'>{args_joined}</span>{end}"
    text = text.replace("\n", "<br />")
    return text


def get_colored_html_text_from_lists(c_list, t_list, sep=""):
    assert len(c_list) == len(  #
        t_list
    ), f"got {len(c_list)} colors and {len(t_list)} texts: {c_list}, {t_list}"
    texts = []
    for c, t in zip(c_list, t_list):
        text = get_colored_html_text(c, t, sep="", end="")
        texts.append(text)
    full_text = sep.join(texts)
    return full_text


def create_html_image_grid(
    pil_images: list[Image.Image],
    texts: Optional[list[str]] = None,
    cell_width: int = 300,
    cell_height: Optional[int] = 500,
):
    pr = NotebookHTMLPrinter()
    pr.open_grid(max_width=cell_width)
    for i, pil_image in enumerate(pil_images):
        if texts is not None:
            text = texts[i]
        else:
            text = ""
        width, height = pil_image.size
        if width > cell_width:
            pil_image = scale_image_to_width(pil_image, 300)
        width, height = pil_image.size
        if cell_height is not None and height > cell_height:
            pil_image = scale_image_to_height(pil_image, 500)
        pr.print(f"<div>{convert_image_to_html(pil_image)}<br>{text}</div>")
        if i < len(pil_images) - 1:
            pr.next_grid_column()
    pr.close_grid()
    return pr.get_html()


def convert_image_to_html(pil_image: Image.Image, addstr: str = "") -> str:
    """
    Usage:
        display(HTML(convert_image_to_html(pil_image)))

    Args:
        pil_image: pillow image object
        addstr: additional string to add to the <img> tag, e.g. "style='width: 50%;'"

    Returns:
        Image as embedded html <img> tag string
    """
    bio = io.BytesIO()
    pil_image.save(bio, "png")
    bios = bio.getbuffer()
    biosb64 = str(base64.b64encode(bios), "ascii")
    addstr = addstr.strip()
    if addstr == "":
        addstr = " "
    else:
        addstr = f" {addstr} "
    html_str = f'<img{addstr}src="data:image/png;base64,{biosb64}"/>'
    return html_str


def convert_figure_to_base64(fig=plt):
    # todo test this if fig is actually not plt
    img = BytesIO()
    fig.savefig(img, format="png")
    img.seek(0)
    plot_url = base64.b64encode(img.getvalue()).decode()
    fig.close()  # Close the plot so that it won't be displayed immediately
    return plot_url


def convert_base64_img_to_tag(base64_img):
    return f'<img src="data:image/png;base64,{base64_img}" />'


def convert_figure_to_html_tag(fig=plt):
    return convert_base64_img_to_tag(convert_figure_to_base64(fig=fig))



import spacy
from spacy.cli import download as spacy_download

SPACY_DEFAULT_EN = "en_core_web_sm"
SPACY_DEFAULT_DE = "de_core_news_lg"


def maybe_download_spacy_model(model=SPACY_DEFAULT_EN):
    try:
        nlp = spacy.load(model)
    except OSError:
        spacy_download(model)
        nlp = spacy.load(model)
    return nlp


loaded_spacy_models = {}


def get_or_maybe_download_spacy_model(model=SPACY_DEFAULT_EN):
    if model in loaded_spacy_models:
        return loaded_spacy_models[model]
    nlp = maybe_download_spacy_model(model)
    loaded_spacy_models[model] = nlp
    return nlp



import pandas as pd
import webbrowser
from contextlib import contextmanager
from loguru import logger
from pandas.io.formats.style import Styler
from pathlib import Path
from typing import Optional

from visiontext.colormaps import redgreen_bright


def create_styler(
    formats=None,
    vmin=0,
    vmax=1,
    cmap=redgreen_bright,
    hide_index=True,
    nan_color="white",
):
    """

    Args:
        formats: dictionary {column name: format string or function}
        vmin: min value for background gradient
        vmax: max value for background gradient
        cmap: colormap for background gradient
        hide_index: hide index column
        nan_color: color for NaN values

    Returns:
        function to use as:
            styled_df = df.style.pipe(styler_fn)
    """
    formats = {} if formats is None else formats

    def styler_fn(
        styler: Styler,
    ):
        styler.format(formats)
        styler.background_gradient(cmap=cmap, axis=1, vmin=vmin, vmax=vmax)
        if hide_index:
            styler.hide(axis="index")
        styler.highlight_null(color=nan_color)

        return styler

    return styler_fn


def save_df_to_html(df, title, html_file, open_browser=True):
    html_file = Path(html_file)
    html_file.parent.mkdir(parents=True, exist_ok=True)
    with html_file.open("w", encoding="utf-8") as f:
        f.write(f"<html><head><title>{title}</title></head><body>")
        f.write(f"<h1>{title}</h1>")
        df.to_html(f)
        f.write("</body></html>")

    if open_browser:
        webbrowser.open(html_file.as_posix())
    logger.info(f"Saved HTML output to {html_file}")


@contextmanager
def full_pandas_display(
    max_rows=None, max_columns=None, max_colwidth=None, width=None, expand_frame_repr=False
):
    """
    A context manager for setting various pandas display options to their
    maximum values so that the output is not truncated when printed.

    Available options: https://pandas.pydata.org/docs/reference/api/pandas.get_option.html

    """
    with pd.option_context(
        "display.max_rows",
        max_rows,
        "display.max_columns",
        max_columns,
        "display.max_colwidth",
        max_colwidth,
        "display.width",
        width,
        "display.expand_frame_repr",
        expand_frame_repr,
    ):
        yield


def pd_print_series(series):
    keys = series.index.tolist()
    values = series.values.tolist()
    keys_str = [str(k) for k in keys]
    values_str = [str(v) for v in values]
    max_key_str_len = max([len(k) for k in keys_str])
    for key, value in zip(keys_str, values_str):
        print(f"{key:{max_key_str_len}}: {value}")


def display_df(
    df, max_rows=None, max_columns=None, max_colwidth=None, width=None, expand_frame_repr=False
):
    with full_pandas_display(max_rows, max_columns, max_colwidth, width, expand_frame_repr):
        print(df)


@contextmanager
def pandas_float_format(fmt="{:.2f}"):
    with pd.option_context("display.float_format", fmt.format):
        yield


def print_stats(input_data, title: Optional[str] = None):
    if title is not None:
        print(title)
    print(pd.Series(input_data).describe())



import os
import requests
import shutil
import zipfile
from PIL import ImageFont

from packg.paths import get_cache_dir


def get_font_path_dejavusans() -> str:
    """
    Find font path of DejaVuSans.ttf or download it to cache folder

    Returns:
        Absolute path to DejaVuSans.ttf
    """
    font_name = "DejaVuSans.ttf"
    font_url = (
        "https://github.com/dejavu-fonts/dejavu-fonts/releases/download/"
        "version_2_37/dejavu-fonts-ttf-2.37.zip"
    )

    try:
        font = ImageFont.truetype(font_name, 14)
        return font.path
        # e.g. /usr/share/fonts/truetype/dejavu/DejaVuSans.ttf
    except OSError:
        pass

    try:
        response = requests.get(font_url)
    except requests.RequestException as e:
        return f"Error downloading the font: {e}"

    response.raise_for_status()  # Check if the download was successful

    font_dir = get_cache_dir() / "fonts"
    font_dir_temp = font_dir / "dejavusans"
    font_archive = font_dir_temp / "archive.zip"

    os.makedirs(font_archive.parent, exist_ok=True)
    with font_archive.open("wb") as f:
        f.write(response.content)

    with zipfile.ZipFile(font_archive.as_posix(), "r") as zip_ref:
        zip_ref.extractall(font_dir_temp / "extracted")

    ttf_src = list(font_dir_temp.glob("**/DejaVuSans.ttf"))[0]
    ttf_dst = font_dir / "DejaVuSans.ttf"
    os.makedirs(ttf_dst.parent, exist_ok=True)
    os.rename(ttf_src, ttf_dst)
    shutil.rmtree(font_dir_temp)
    return ttf_dst.as_posix()


def main():
    print(get_font_path_dejavusans())


if __name__ == "__main__":
    main()



"""
Original source: https://github.com/lpryszcz/tar_indexer
LICENSE is GNUGPLV3.0

Author:
    l.p.pryszcz@gmail.com
    Mizerow, 23/04/2014
Changes:
    - 2to3
    - remove script usage
    - small tweaks

TAR indexer uses sqlite3 for index storing and allows indexing of multiple tar archives.
Note, only raw (uncompressed) tar files are accepted as native tar.gz cannot be random accessed.
But you can compress each file using zlib before adding it to tar.

Prerequisites
    Python 3.7+
    sqlite3 - install with sudo apt-get install / with conda

"""

import os
import sqlite3
import tarfile
import time
from pathlib import Path


def get_cursor(indexfn):
    """Return cur object"""
    # create/connect to sqlite3 database
    cnx = sqlite3.connect(indexfn)
    cur = cnx.cursor()
    # asyn execute >50x faster ##http://www.sqlite.org/pragma.html#pragma_synchronous
    # cur.execute("PRAGMA synchronous=OFF")
    # prepare db schema #type='table' and
    cur.execute("SELECT * FROM sqlite_master WHERE name='file_data'")
    result = cur.fetchone()
    if not result:
        # create file data table
        cur.execute(
            """CREATE TABLE file_data (file_id INTEGER PRIMARY KEY,
                    file_name TEXT, mtime FLOAT)"""
        )
        cur.execute("CREATE INDEX file_data_file_name ON file_data (file_name)")
        # create offset_data table
        cur.execute(
            """CREATE TABLE offset_data (file_id INTEGER,
                    file_name TEXT, offset INTEGER, file_size INTEGER)"""
        )  # ,PRIMARY KEY (file_id, file_name))
        cur.execute("CREATE INDEX offset_data_file_id ON offset_data (file_id)")
        cur.execute("CREATE INDEX offset_data_file_name ON offset_data (file_name)")
    return cur


def prepare_db(base_path, cur, tarfn_str, verbose):
    """Prepare database and add file"""
    tarfn_full = Path(base_path) / tarfn_str
    mtime = os.path.getmtime(tarfn_full)
    # check if file already indexed
    cur.execute("SELECT file_id, mtime FROM file_data WHERE file_name = ?", (tarfn_str,))
    result = cur.fetchone()  # ; print result
    if result:
        file_id, pmtime = result  # ; print file_id, pmtime
        # skip if index with newer mtime than archives exists
        if mtime <= pmtime:
            if verbose:
                print(" Archive already indexed.")
            return None
        # else update mtime and remove previous index
        cur.execute("UPDATE file_data SET mtime = ? WHERE file_name = ?", (mtime, tarfn_str))
        cur.execute("DELETE FROM offset_data WHERE file_id = ?", (file_id,))
    else:
        cur.execute("SELECT MAX(file_id) FROM file_data")
        (max_file_id,) = cur.fetchone()  # ; print max_file_id
        if max_file_id:
            file_id = max_file_id + 1
        else:
            file_id = 1
        # add file info
        cur.execute("INSERT INTO file_data VALUES (?, ?, ?)", (file_id, tarfn_str, mtime))
    return file_id


def index_tar(base_path, tarfn, indexfn, verbose=True, sleep_duration=5):
    """Index tar file and create sqlite index."""
    if verbose:
        print(f"Indexing tarfile {base_path} / {tarfn} to {indexfn}")
    if not indexfn:
        raise ValueError(f"Index missing {indexfn}")
    # get archive size
    tarfn_str = str(tarfn)
    tarfn_full = Path(base_path) / tarfn
    if not tarfn_full.is_file():
        raise FileNotFoundError(f"File not found: {tarfn_full}")
    tarsize = os.path.getsize(tarfn_full)
    # prepare db
    cur = get_cursor(indexfn)
    file_id = prepare_db(base_path, cur, tarfn_str, verbose)
    if not file_id:
        return
    tar = tarfile.open(tarfn_full)
    data = []
    i = 1
    n_files = 0
    for tarinfo in tar:
        if not tarinfo.isfile():
            continue
        data.append((file_id, tarinfo.name, tarinfo.offset_data, tarinfo.size))
        n_files += 1
        # upload
        if i % 100 == 0:
            cur.executemany("INSERT INTO offset_data VALUES (?, ?, ?, ?)", data)
            if verbose:
                print(f" {i} [{tarinfo.offset_data / tarsize:.2%}]      ", end="\r")
            # free ram...
            data = []
            tar.members = []
        i += 1
    # upload last batch
    if len(data) > 0:
        cur.executemany("INSERT INTO offset_data VALUES (?, ?, ?, ?)", data)
    # finally commit changes
    cur.connection.commit()
    print(f"Wrote {n_files} files to {indexfn}. Sleep {sleep_duration}s to let the DB work...")
    time.sleep(sleep_duration)



from omegaconf import OmegaConf

from typedparser.objects import modify_nested_object


def load_dotlist(merge_dotlist: list[str]) -> dict:
    """Converts a list of key=value strings into a nested dictionary with dot notation support.

    Type conversions: int, float, bool, comma-separated lists, empty values become None.

    Args:
        merge_dotlist: List of strings in the format "key=value" or "parent.child=value".

    Returns:
        A dictionary with the parsed key-value pairs, with appropriate type conversions.
    """
    conf_dotlist = OmegaConf.from_dotlist(merge_dotlist)
    dict_dotlist = OmegaConf.to_container(conf_dotlist, resolve=True)

    def _convert_leaf(val):
        # convert commas to lists to allow for list overrides
        if isinstance(val, str):
            if "," in val:
                val = val.split(",")
                out_val = []
                for i, v in enumerate(val):
                    v = v.strip()
                    if v == "" and i == len(val) - 1:
                        # skip last empty string, so "key=value," creates the list ["value"].
                        continue
                    out_val.append(int(v) if v.isdigit() else v)
                val = out_val
        return val

    modify_nested_object(dict_dotlist, _convert_leaf)

    return dict_dotlist



"""
Check whether GPU is available for PyTorch or TensorFlow.

Oneliners:

python -c "from tensorflow.python.client import device_lib; print(device_lib.list_local_devices())"
python -c "import torch; print(torch.cuda.is_available()); print(torch.cuda.device_count()); print(torch.version.cuda)"

Torch backends: https://pytorch.org/docs/stable/backends.html

"""

from packg import format_exception
from visiontext.profiling import get_gpu_profiler, GPUProfilerTorch


def main():
    print(f"\nCheck PyTorch GPU availability")
    try:
        import torch  # noqa

        print(f"    torch.version.__version__={torch.version.__version__}")
        print(f"    torch.version.cuda={torch.version.cuda}")
        print(f"    torch.cuda.is_available()={torch.cuda.is_available()}")
        print(f"    torch.cuda.device_count()={torch.cuda.device_count()}")
        print(f"    torch.backends.cudnn.version()={torch.backends.cudnn.version()}")
        print(f"    torch.cuda.get_device_capability()={torch.cuda.get_device_capability()}")
        print(f"    torch.cuda.is_bf16_supported()={torch.cuda.is_bf16_supported()}")
        # https://github.com/pytorch/pytorch/issues/75427
        # before compute capability 8.0 bfloat16 does not offer speed-ups and might fail.
        example_tensor = torch.zeros(128, 1024, 768).cuda()
        print(f"    example_tensor.shape={example_tensor.shape}")

    except Exception as e:
        print(f"    PyTorch failed: {format_exception(e)}")

    print(f"\nCheck TensorFlow GPU availability")
    try:
        from tensorflow.python.client import device_lib  # noqa

        print(device_lib.list_local_devices())
    except Exception as e:
        print(f"    TensorFlow failed: {format_exception(e)}")

    profiler = get_gpu_profiler()
    print(profiler)
    print(f"\nGPU Profiler output: {profiler.profile_to_str()}")


if __name__ == "__main__":
    main()



"""
original nltk tokenizer lose the space information.

here the space information is preserved and the functions outputs list of tuples like
[['skiing', ' '], ['on', ' '], ['a', ' '], ['mountain', ' '], ['slope', ''], ['.', '']]

which can be used to reconstruct the original text input.

"""

from __future__ import annotations

import itertools

import nltk
from nltk.data import load
from nltk.tokenize.destructive import NLTKWordTokenizer


def ensure_setup_nltk(language="english"):
    try:
        load(f"tokenizers/punkt/{language}.pickle")
    except LookupError:
        nltk.download("punkt", quiet=True)


def tokenize_sentences_and_connectors(text: str, language="english") -> list[tuple[str, str]]:
    tokenizer = load(f"tokenizers/punkt/{language}.pickle")  # nltk caches internally
    return apply_nltk_tokenizer(tokenizer, text)


_treebank_word_tokenizer = None


def get_treebank_word_tokenizer():
    global _treebank_word_tokenizer
    if _treebank_word_tokenizer is None:
        _treebank_word_tokenizer = NLTKWordTokenizer()
    return _treebank_word_tokenizer


def tokenize_words_and_connectors(
    text: str, language="english", split_eos=True
) -> list[tuple[str, str]]:
    """
    Split text into tokens and remember all spaces so the text can be reconstructed.

    Args:
        text: input text
        language: tokenizer language
        split_eos: if true (default) eos tokens like "." will be returned as a separate list item.
            if false they will be appended to the previous word.

    Returns:
        list of tuples (word, connector) that can reconstruct the original text:
            word1 + connector1 + word2 + connector2 + ... = text
    """

    sentences_and_connectors = (
        [(text, "")] if not split_eos else tokenize_sentences_and_connectors(text, language)
    )
    words_and_connectors = []
    for sentence, sentence_connector in sentences_and_connectors:
        words_conns_here = apply_nltk_tokenizer(get_treebank_word_tokenizer(), sentence)
        for iw, (word, connector) in enumerate(words_conns_here):
            if iw == len(words_conns_here) - 1:
                connector = f"{connector}{sentence_connector}"
            words_and_connectors.append([word, connector])
    return words_and_connectors


def rebuild_from_words_and_connectors(words_and_connectors: list[tuple[str, str]]) -> str:
    return "".join(itertools.chain(*words_and_connectors))


def apply_nltk_tokenizer(tokenizer, text: str) -> list[tuple[str, str]]:
    """

    Args:
        tokenizer: nltk tokenizer
        text: input text

    Returns: list of tuples (word, connector) that can reconstruct the original text:
        word1 + connector1 + word2 + connector2 + ... = text

    """
    tokenizer_output = list(tokenizer.span_tokenize(text))
    if len(tokenizer_output) == 0:
        return []
    starts, stops = (list(a) for a in zip(*tokenizer_output))
    starts.append(len(text))

    output = []
    for n in range(len(stops)):
        word = text[starts[n] : stops[n]]
        connector = text[stops[n] : starts[n + 1]]
        output.append((word, connector))
    return output


def main():
    inp = """\
The image features two people standing on a snow-covered slope, both wearing skis and holding ski \
poles. They are posing for a picture, likely enjoying their time skiing on the mountain. The \
individuals are standing next to each other, with one person on the left and the other on the \
right. Their skis are placed in various positions, with one set of skis extending horizontally \
and the other set angled upward. The scene captures the excitement and fun of skiing on a \
mountain slope.\
"""
    print(inp)
    words_and_connectors = tokenize_words_and_connectors(inp)
    print(words_and_connectors)
    words_and_connectors2 = tokenize_words_and_connectors(inp, split_eos=True)
    print(words_and_connectors2)
    for w1c1, w2c2 in zip(words_and_connectors, words_and_connectors2):
        if w1c1 != w2c2:
            print(w1c1, "========|========", w2c2)

    reconstruct = "".join(itertools.chain(*words_and_connectors))
    print(reconstruct == inp)
    print()

    sentences_and_connectors = tokenize_sentences_and_connectors(inp)
    print(sentences_and_connectors)
    reconstruct = "".join(itertools.chain(*sentences_and_connectors))
    print(reconstruct == inp)
    # todo add tests, also test things like "" and "\n" as they can break things


if __name__ == "__main__":
    main()



import seaborn as sns
from matplotlib import pyplot as plt


def plot_hist(data, bins=None, density=False, title="untitled", figsize=None, show=False):
    plt.figure(figsize=figsize)
    plt.title(title)
    plt.hist(data, density=density, bins=bins)
    sns.kdeplot(data, bw_adjust=1)
    plt.grid()
    if show:
        plt.show()



from pathlib import Path
from typing import (
    Optional,
)

from attrs import define
from datasets import load_dataset
from loguru import logger

from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from packg.paths import get_data_dir
from typedparser import VerboseQuietArgs, add_argument, TypedParser


@define
class Args(VerboseQuietArgs):
    base_dir: Optional[Path] = add_argument(
        shortcut="-d", type=str, help="Base dir", default=get_data_dir()
    )
    dataset_name: str = add_argument(positional=True, type=str, help="Dataset name")
    options: list[str] | None = add_argument(shortcut="-o", action="append", help="dataset kwargs")


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    dict_dotlist = load_dotlist(args.options)  # TODO ADD TO VISIONTEXT
    # takes ~26 min
    dataset = load_dataset(
        args.dataset_name,
        use_auth_token=True,  # required
        streaming=False,  # this downloads to disk
        split="train",
        cache_dir=(args.base_dir / "huggingface-en-2301").as_posix(),
    )


if __name__ == "__main__":
    main()



import importlib
import importlib.util
import shutil
import sys


def main():
    lib_module_name = __file__.replace("\\", "/").split("/")[-2]
    if len(sys.argv) <= 1:
        print(f"Usage: {lib_module_name} <module> <*args>\n")
        from packg.packaging import get_modules_for_autocomplete
        from packg.strings import format_pseudo_table

        output_modules = get_modules_for_autocomplete(lib_module_name, run_dir=None)
        terminal_width = shutil.get_terminal_size().columns
        print(format_pseudo_table(output_modules, max_width=terminal_width))
        sys.exit(1)
    target_module = sys.argv[1]
    args = sys.argv[2:]
    # load the spec to modify sys.argv before importing, otherwise it's too late
    module_to_load = f"{lib_module_name}.{target_module}"
    spec = importlib.util.find_spec(module_to_load)
    if spec is None:
        raise ValueError(f"Module {module_to_load} not found")
    origin = spec.origin
    sys.argv = [origin] + args
    # import the module and run the main function
    module = importlib.import_module(module_to_load)
    module.main()


if __name__ == "__main__":
    main()



"""
Collection of useful imports e.g. for jupyter notebooks. Usage:

>>> from visiontext.imports import *

"""

from collections import defaultdict, Counter

import base64
import hashlib
import io
import json
import numpy as np
import os
import pandas as pd
import pickle
import random
import re
import shutil
import sys
import time
from IPython.display import (
    Image,
    display,
    HTML,
    JSON,
)
from PIL import Image
from copy import deepcopy
from enum import Enum
from loguru import logger
from matplotlib import pyplot as plt
from natsort import natsorted
from pathlib import Path
from pprint import pprint, pformat
from timeit import default_timer as timer
from timeit import default_timer
from tqdm import tqdm
from typing import (
    Dict,
    List,
    Optional,
    Any,
    Iterable,
    Mapping,
    Tuple,
    Union,
    Callable,
    BinaryIO,
    Sequence,
    Collection,
)

from packg import format_exception
from packg.iotools.jsonext import load_json, dump_json, loads_json, dumps_json
from packg.log import configure_logger
from packg.magic import reload_recursive
from packg.paths import get_cache_dir, get_data_dir
from packg.strings import b64_encode_from_bytes
from packg.tqdmext import tqdm_max_ncols
from typedparser.objects import invert_dict_of_dict, flatten_dict
from visiontext.htmltools import NotebookHTMLPrinter, display_html_table
from visiontext.images import PILImageScaler, open_image_scaled
from visiontext.nlp.regextools import preprocess_text_simple
from visiontext.pandatools import full_pandas_display

# the __all__ list below is used to stop pycharm or other tools from removing unused imports
# to update it after changing the imports above, uncomment the code below and copypaste the output
# python -c "from visiontext.imports import *; im = [m for m in globals().keys() if not m.startswith(\"_\")]; print(f\"__all__ = {repr(sorted(im))}\")"
__all__ = [
    "Any",
    "BinaryIO",
    "Callable",
    "Collection",
    "Counter",
    "Dict",
    "Enum",
    "HTML",
    "Image",
    "Iterable",
    "List",
    "Mapping",
    "NotebookHTMLPrinter",
    "Optional",
    "PILImageScaler",
    "Path",
    "Sequence",
    "Tuple",
    "Union",
    "b64_encode_from_bytes",
    "base64",
    "configure_logger",
    "deepcopy",
    "defaultdict",
    "display",
    "display_html_table",
    "dump_json",
    "dumps_json",
    "flatten_dict",
    "format_exception",
    "full_pandas_display",
    "get_cache_dir",
    "get_data_dir",
    "hashlib",
    "invert_dict_of_dict",
    "io",
    "json",
    "load_json",
    "loads_json",
    "logger",
    "np",
    "open_image_scaled",
    "os",
    "pd",
    "pformat",
    "pickle",
    "plt",
    "pprint",
    "random",
    "re",
    "reload_recursive",
    "shutil",
    "sys",
    "time",
    "timer",
    "default_timer",
    "tqdm",
    "tqdm_max_ncols",
    "JSON",
    "preprocess_text_simple",
    "natsorted",
]



"""
Image utilities using opencv-python, libturbojpeg, pillow

which sampling method to use:
    - bilinear is a classic
    - area might be best for downsampling
    - nearest for upsampling without blurring
    - cubic/lanczos might look better for upsampling but are slower

https://www.libjpeg-turbo.org/
documentation on tjCompress2 method:
https://rawcdn.githack.com/libjpeg-turbo/libjpeg-turbo/main/doc/html/group___turbo_j_p_e_g.html

conda install -c conda-forge libjpeg-turbo -y
pip install -U pyturbojpeg opencv-python pillow

Todo: Write tests, does it work with all kinds of images (grayscale, RGB, RGBA)

Todo: this CMYK file breaks libjpeg-turbo: datasets/imagenet1k/val/n13133613/ILSVRC2012_val_00019877.JPEG

Examples:
    >>> from IPython.display import display
    >>> image = open_image_scaled("image.png", bigger_side=500)
    >>> display(image)
"""

from __future__ import annotations

import io
import numpy as np
import os
import torch
from PIL import Image
from PIL.Image import Image as PILImage, Resampling
from dataclasses import dataclass
from tempfile import NamedTemporaryFile
from typing import Union, Optional

from packg import format_exception
from packg.constclass import Const


class SamplingConst:
    NEAREST = "nearest"
    BILINEAR = "bilinear"
    BICUBIC = "bicubic"
    LANCZOS = "lanczos"
    AREA = "area"
    AUTO = "auto"


# optional packages opencv-python and libturbojpeg

try:
    import cv2
except ImportError:
    cv2 = None


try:
    import turbojpeg
except ImportError:
    turbojpeg = None

SamplingMapPIL = {
    SamplingConst.NEAREST: Resampling.NEAREST,
    SamplingConst.BILINEAR: Resampling.BILINEAR,
    SamplingConst.BICUBIC: Resampling.BICUBIC,
    SamplingConst.LANCZOS: Resampling.LANCZOS,
    SamplingConst.AREA: Resampling.BOX,
    "box": Resampling.BOX,  # backward compatibility
    "hamming": Resampling.HAMMING,
}
DEFAULT_UPSAMPLING_METHOD = SamplingConst.BICUBIC
DEFAULT_DOWNSAMPLING_METHOD = SamplingConst.AREA
JPEG_QUAL_FFMPEG = 2  # 2 best, 31 worst
AUTO = "auto"

ImageType = Union[np.ndarray, Image.Image]
ResamplingMethodType = Union[str, int, Resampling]


class JPEGDecoderConst(Const):
    OPENCV = "opencv"
    LIBTURBOJPEG_DEFAULT = "libturbojpeg_default"
    LIBTURBOJPEG_FASTEST = "libturbojpeg_fastest"
    PILLOW = "pillow"
    PILLOW_IMAGE = "pillow_image"  # returns PIL.Image instead of np.ndarray


class _TjpegGetter:
    def __init__(self):
        self.jpeg = None

    def get(self):
        if self.jpeg is None:
            if turbojpeg is None:
                raise ImportError(
                    "turbojpeg not installed. To install, run:\n"
                    "conda install -c conda-forge libjpeg-turbo -y\n"
                    "pip install -U pyturbojpeg"
                )
            self.jpeg = turbojpeg.TurboJPEG()
        return self.jpeg


_jpeg_getter = _TjpegGetter()


def get_dummy_image_np() -> np.array:
    """
    Returns:
        numpy array image type uint8, shape (100, 100, 3), y-axis gradient from black to white
    """
    return (
        np.linspace(0, 255, num=100)
        .astype(np.uint8)[:, None, None]
        .repeat(3, axis=2)
        .repeat(100, axis=1)
    )


def check_turbojpeg_available() -> bool:
    if turbojpeg is None:
        return False
    try:
        test_method = JPEGDecoderConst.LIBTURBOJPEG_DEFAULT
        image_in = get_dummy_image_np()
        jpeg_bytes = encode_jpeg(image_in, method=test_method)
        image_out = decode_jpeg(jpeg_bytes, method=test_method)
        np.testing.assert_allclose(image_in, image_out, atol=1, rtol=1e-3)
    except Exception as e:
        print(f"Error in turbojpeg test: {format_exception(e)}")
        return False
    return True


def decode_jpeg(
    jpeg_arr: Union[np.ndarray, bytes],
    method: str = JPEGDecoderConst.LIBTURBOJPEG_DEFAULT,
    is_gray: bool = False,
) -> Union[np.ndarray, PILImage]:
    """
    Decode jpeg encoded image.
    Don't care about autodetecting gray jpegs, probably not worth it.

    Args:
        jpeg_arr: input jpeg encoded image shape (num_bytes,)
        method: decoder
        is_gray: if true, decode to grayscale

    Returns:
        decoded image shape (h, w, 3) dtype uint8 in [0, 255] OR pillow image
    """
    decoded_arr = None
    if method == JPEGDecoderConst.OPENCV:
        import cv2

        if isinstance(jpeg_arr, bytes):
            jpeg_arr = np.frombuffer(jpeg_arr, dtype=np.uint8)
        if is_gray:
            decoded_arr = cv2.imdecode(jpeg_arr, cv2.IMREAD_GRAYSCALE)
        else:
            decoded_arr = cv2.imdecode(jpeg_arr, cv2.IMREAD_COLOR)
            decoded_arr = cv2.cvtColor(decoded_arr, cv2.COLOR_BGR2RGB)
    if (
        method == JPEGDecoderConst.LIBTURBOJPEG_DEFAULT
        or method == JPEGDecoderConst.LIBTURBOJPEG_FASTEST
    ):
        import turbojpeg

        flags = 0
        if method == JPEGDecoderConst.LIBTURBOJPEG_FASTEST:
            flags |= turbojpeg.TJFLAG_FASTUPSAMPLE | turbojpeg.TJFLAG_FASTDCT
        try:
            decoded_arr = _jpeg_getter.get().decode(
                jpeg_arr,
                pixel_format=turbojpeg.TJPF_GRAY if is_gray else turbojpeg.TJPF_RGB,
                flags=flags,
            )
        except OSError as e:
            # it seems there exist images that cannot be decoded correctly with libjpeg-turbo
            # but pillow decodes them perfectly fine. therefore catch those errors and call pillow.
            error_str = format_exception(e)
            if (
                error_str == "OSError: Unsupported color conversion request"
                or "Could not determine subsampling type for JPEG image" in error_str
            ):
                # logger.error(
                #     f"{error_str} for image bytes of length {len(jpeg_arr)}. "
                #     f"Decoding with pillow instead of libjpeg-turbo."
                # )
                method = JPEGDecoderConst.PILLOW
                decoded_arr = None
            else:
                raise e
    if method == JPEGDecoderConst.PILLOW:
        # noinspection PyTypeChecker
        decoded_arr = np.array(Image.open(io.BytesIO(jpeg_arr)).convert("L" if is_gray else "RGB"))
    if method == JPEGDecoderConst.PILLOW_IMAGE:
        pil_image = Image.open(io.BytesIO(jpeg_arr)).convert("L" if is_gray else "RGB")
        return pil_image
    if decoded_arr is None:
        raise ValueError(f"Unknown JPEG decoding method {method}")
    if is_gray and decoded_arr.ndim == 3:
        decoded_arr = np.squeeze(decoded_arr, axis=-1)
    return decoded_arr


def encode_jpeg(
    np_arr: Union[np.ndarray, PILImage],
    method: str = JPEGDecoderConst.LIBTURBOJPEG_DEFAULT,
    quality=95,
) -> bytes:
    """
    Encode image to jpeg. expects (h, w) or (h, w, 1) for gray-scale and
    (h, w, 3) for RGB images.

    Compared losses and speed for different combinations of en-/decoder and quality
    both with random noise and some real images b/c noise probably cant be encoded well.
    Result: libturbojpeg_fastest has too many artifacts.
    libturbojpeg_default is fast and good quality, use that.
    Others are slower and not better quality-wise.

    quality: 95 best and cv2 default, 75 pillow default, 85 libturbojpeg default, 0-31 worst
    """
    is_gray = None
    if method == JPEGDecoderConst.PILLOW_IMAGE:
        is_gray = np_arr.mode == "L"
    elif np_arr.ndim == 2:
        is_gray = True
    elif np_arr.ndim == 3:
        if np_arr.shape[-1] == 1:
            is_gray = True
            np_arr = np.squeeze(np_arr, axis=2)
        elif np_arr.shape[-1] == 3:
            is_gray = False

    if is_gray is None:
        raise ValueError("Unknown image shape: " + str(np_arr.shape))

    if method == JPEGDecoderConst.OPENCV:
        import cv2

        params = (cv2.IMWRITE_JPEG_QUALITY, quality)
        if is_gray:
            encoded_arr = cv2.imencode(".jpg", np_arr, params=params)[1]
        else:
            np_arr_bgr = cv2.cvtColor(np_arr, cv2.COLOR_RGB2BGR)
            encoded_arr = cv2.imencode(".jpg", np_arr_bgr, params=params)[1]
        # creates np.ndarray with shape (num_bytes) dtype uint8
        return encoded_arr.tobytes()
    if method == JPEGDecoderConst.PILLOW:
        mode = "L" if is_gray else "RGB"
        bio = io.BytesIO()
        # noinspection PyTypeChecker
        Image.fromarray(np_arr, mode).save(bio, format="JPEG", quality=quality)
        return bio.getvalue()
    if method == JPEGDecoderConst.PILLOW_IMAGE:
        bio = io.BytesIO()
        # noinspection PyTypeChecker
        np_arr.save(bio, format="JPEG", quality=quality)
        return bio.getvalue()
    if method == JPEGDecoderConst.LIBTURBOJPEG_DEFAULT:
        import turbojpeg
        import cv2

        np_arr_bgr = cv2.cvtColor(np_arr, cv2.COLOR_RGB2BGR)
        encoded_arr = _jpeg_getter.get().encode(
            np_arr_bgr, quality=quality, **_get_tjpeg_kwargs(is_gray)
        )
        return encoded_arr
    if method == JPEGDecoderConst.LIBTURBOJPEG_FASTEST:
        import turbojpeg
        import cv2

        np_arr_bgr = cv2.cvtColor(np_arr, cv2.COLOR_RGB2BGR)
        encoded_arr = _jpeg_getter.get().encode(
            np_arr_bgr,
            quality=quality,
            flags=turbojpeg.TJFLAG_FASTUPSAMPLE | turbojpeg.TJFLAG_FASTDCT,
            **_get_tjpeg_kwargs(is_gray),
        )
        return encoded_arr
    raise ValueError(f"Unknown JPEG encoding method {method}")


def _get_tjpeg_kwargs(is_gray: bool) -> dict[str, int]:
    return {
        "pixel_format": turbojpeg.TJPF_GRAY if is_gray else turbojpeg.TJPF_BGR,
        "jpeg_subsample": turbojpeg.TJSAMP_GRAY if is_gray else turbojpeg.TJSAMP_422,
    }


@dataclass
class PILImageScaler:
    """
    Scale an image using PIL.

    todo in a new version set the default to return pillow and
        return_type as enum with pillow, uint8, fp32
    """

    upsampling_method: Union[str, int] = DEFAULT_UPSAMPLING_METHOD
    downsampling_method: Union[str, int] = DEFAULT_DOWNSAMPLING_METHOD
    return_pillow: bool = False
    return_fp32: bool = False

    def scale_image_smaller_side(
        self, img: ImageType, smaller_side: int, method: ResamplingMethodType = AUTO
    ):
        img = self._ensure_pil(img)
        w, h = img.size
        if h < w:
            target_h = smaller_side
            target_w = round(w * smaller_side / h)
        else:
            target_w = smaller_side
            target_h = round(h * smaller_side / w)
        method = self._get_method(h, w, target_h, target_w, method)
        img = img.resize((target_w, target_h), method)
        return self._prepare_return(img)

    def scale_image_bigger_side(
        self, img: ImageType, bigger_side: int, method: ResamplingMethodType = AUTO
    ):
        img = self._ensure_pil(img)
        w, h = img.size
        if h > w:
            target_h = bigger_side
            target_w = int(w * bigger_side / h)
        else:
            target_w = bigger_side
            target_h = int(h * bigger_side / w)
        method = self._get_method(h, w, target_h, target_w, method)
        img = img.resize((target_w, target_h), method)
        return self._prepare_return(img)

    def scale_image_to_width(
        self, img: ImageType, target_w: int, method: ResamplingMethodType = AUTO
    ):
        img = self._ensure_pil(img)
        w, h = img.size
        target_h = int(h * target_w / w)
        method = self._get_method(h, w, target_h, target_w, method)
        img = img.resize((target_w, target_h), method)
        return self._prepare_return(img)

    def scale_image_to_height(
        self, img: ImageType, target_h: int, method: ResamplingMethodType = AUTO
    ):
        img = self._ensure_pil(img)
        w, h = img.size
        target_w = int(w * target_h / h)
        method = self._get_method(h, w, target_h, target_w, method)
        img = img.resize((target_w, target_h), method)
        return self._prepare_return(img)

    def scale_image(
        self, img: ImageType, target_h: int, target_w: int, method: ResamplingMethodType = AUTO
    ):
        img = self._ensure_pil(img)
        w, h = img.size
        method = self._get_method(h, w, target_h, target_w, method)
        img = img.resize((target_w, target_h), method)
        return self._prepare_return(img)

    def crop_square(self, img: ImageType):
        img = self._ensure_pil(img)
        w, h = img.size
        if w == h:
            return img
        if w > h:
            left = (w - h) // 2
            right = left + h
            top = 0
            bottom = h
        else:
            top = (h - w) // 2
            bottom = top + w
            left = 0
            right = w
        img = img.crop((left, top, right, bottom))
        return self._prepare_return(img)

    def _ensure_pil(self, img: ImageType):
        if isinstance(img, Image.Image):
            return img
        return Image.fromarray(img)

    def _get_method(
        self, h: int, w: int, target_h: int, target_w: int, method: ResamplingMethodType = AUTO
    ) -> Resampling:
        if isinstance(method, str):
            method = method.lower()
        if method == AUTO:
            method = (
                self.downsampling_method
                if _is_downsampling(h, w, target_h, target_w)
                else self.upsampling_method
            )
        if isinstance(method, str):
            method = SamplingMapPIL[method]
        return method

    def _prepare_return(self, img: ImageType):
        if self.return_pillow:
            return img
        img = np.array(img)
        if self.return_fp32:
            img = img.astype(np.float32) / 255.0
            img = np.clip(img, 0.0, 1.0)
        return img


pil_editor = PILImageScaler(return_pillow=True)

scale_image_smaller_side = pil_editor.scale_image_smaller_side
scale_image_bigger_side = pil_editor.scale_image_bigger_side
scale_image_to_width = pil_editor.scale_image_to_width
scale_image_to_height = pil_editor.scale_image_to_height
scale_image = pil_editor.scale_image
crop_image_square = pil_editor.crop_square


def _is_downsampling(h, w, target_h, target_w):
    return target_h <= h and target_w <= w


def open_image_scaled(
    image_file,
    smaller_side: Optional[int] = None,
    bigger_side: Optional[int] = None,
    method: str = AUTO,
    upsampling_method: str = DEFAULT_UPSAMPLING_METHOD,
    downsampling_method: str = DEFAULT_DOWNSAMPLING_METHOD,
    convert: Optional[str] = None,  # RGB, L, ...
):
    scaler = PILImageScaler(
        upsampling_method=upsampling_method,
        downsampling_method=downsampling_method,
        return_pillow=True,
    )
    image = Image.open(image_file)
    if convert is not None:
        image = image.convert(convert)
    if smaller_side is not None and bigger_side is not None:
        raise ValueError(
            f"Only one of smaller_side={smaller_side} and bigger_side={bigger_side} can be given"
        )
    elif smaller_side is not None:
        return scaler.scale_image_smaller_side(image, smaller_side, method=method)
    elif bigger_side is not None:
        return scaler.scale_image_bigger_side(image, bigger_side, method=method)
    else:
        raise ValueError(
            f"One of smaller_side={smaller_side} and bigger_side={bigger_side} must be given"
        )


def get_image_properties(image_file):
    """
    Modes: https://pillow.readthedocs.io/en/stable/handbook/concepts.html

    Args:
        image_file:

    Returns:

    """
    image = Image.open(image_file)
    mode = image.mode  # L (grayscale), RGB, RGBA
    assert mode in ["L", "RGB", "RGBA"], f"Unsupported mode {mode} for {image_file}"
    width, height = image.size
    return mode, width, height


def show_image_pil(image: ImageType):
    if isinstance(image, np.ndarray):
        image = Image.fromarray(image)
    image.show()


def get_interpolation_for_cv2(method: str):
    # ref: https://chadrick-kwag.net/cv2-resize-interpolation-methods/
    import cv2

    sampling_map_cv2 = {
        SamplingConst.NEAREST: cv2.INTER_NEAREST,
        SamplingConst.BILINEAR: cv2.INTER_LINEAR,
        SamplingConst.AREA: cv2.INTER_AREA,
        SamplingConst.BICUBIC: cv2.INTER_CUBIC,
        SamplingConst.LANCZOS: cv2.INTER_LANCZOS4,
    }
    if method not in sampling_map_cv2:
        raise KeyError(
            f"Unknown image scaling method for cv2: {method}, "
            f"available methods: {list(sampling_map_cv2.keys())}"
        )
    return sampling_map_cv2[method]


def get_pil_image_from_torch_tensor(tensor: torch.Tensor) -> PILImage:
    """
    Convert a torch tensor to a PIL image.

    Args:
        tensor: tensor with shape (C, H, W) or (B, C, H, W)

    Returns:
        PIL image
    """
    if tensor.ndim == 4:
        assert tensor.shape[0] == 1, f"Expected batch size 1, got {tensor.shape[0]}"
        tensor = tensor[0]
    tensor = tensor.detach().cpu().numpy()
    tensor = np.moveaxis(tensor, 0, -1)
    tensor = np.clip(tensor, 0.0, 1.0)
    tensor = (tensor * 255).astype(np.uint8)
    return Image.fromarray(tensor)


def display_image_using_kitten_icat(image: Image.Image):
    tempf = NamedTemporaryFile(delete=False, suffix=".jpg")
    image.save(tempf.name)
    tempf.close()
    os.system(f'kitten icat "{tempf.name}"')
    os.unlink(tempf.name)



from __future__ import annotations

import datetime
import h5py
import os
from attr import define, field
from loguru import logger
from pathlib import Path
from spacy import Language

from packg.paths import get_data_dir
from packg.strings import quote_with_urlparse
from visiontext.nlp.spacytools import maybe_download_spacy_model, SPACY_DEFAULT_EN


class LemmatizerInterface:
    def lemmatize(self, in_str: str) -> list[str]:
        raise NotImplementedError

    def batch_lemmatize(self, sentences: list[str]) -> list[list[str]]:
        raise NotImplementedError

    def get_unique_name(self) -> str:
        raise NotImplementedError


@define(slots=False)
class LemmatizerSpacy(LemmatizerInterface):
    name: str = SPACY_DEFAULT_EN
    _lemmatizer: Language | None = field(init=False, repr=False, default=None)
    verbose: bool = False

    def get_unique_name(self) -> str:
        return f"spacy_{self.name}"

    @property
    def lemmatizer(self):
        if self._lemmatizer is None:
            self._lemmatizer = maybe_download_spacy_model(self.name)
        return self._lemmatizer

    def lemmatize(self, in_str: str) -> list[str]:
        doc = self.lemmatizer(in_str)
        words_out = []
        for token in doc:
            if token.pos_ in ["NOUN", "VERB"]:
                words_out.append(token.lemma_)
            else:
                words_out.append(token.text)
        return words_out

    def batch_lemmatize(self, sentences: list[str]) -> list[list[str]]:
        sentences_set = set(sentences)
        missing_sentences = sentences_set
        if self.verbose:
            logger.info(f"Lemmatizing {len(missing_sentences)} sentences")
        output_dict = {}
        for sentence in missing_sentences:
            words = self.lemmatize(sentence)
            output_dict[sentence] = words
        output_list = [output_dict[sentence] for sentence in sentences]
        return output_list


SEP_CHAR = "\x00"


@define(slots=False)
class LemmatizerDbWrapper(LemmatizerInterface):
    """
    Wrap a database (h5) to store the results for each sentence.
    """

    lemmatizer: LemmatizerInterface
    compute_missing: bool = True
    save_to_db: bool = True
    h5_file: Path = None

    def __attrs_post_init__(self):
        self.h5_file = (
            get_data_dir() / f"text_embeddings/lemmas/{self.lemmatizer.get_unique_name()}"
            if self.h5_file is None
            else self.h5_file
        )

    def lemmatize(self, in_str: str) -> list[str]:
        return self.batch_lemmatize([in_str])[0]

    def batch_lemmatize(self, sentences: list[str]) -> list[list[str]]:
        missing_sentences_set = set(sentences)

        output_dict = {}
        if self.h5_file.is_file():
            # read existing embeddings
            with h5py.File(self.h5_file, "r", libver="latest", swmr=True) as f:
                for sentence in list(missing_sentences_set):
                    quoted_sentence = quote_with_urlparse(sentence, prefix="q")
                    if quoted_sentence in f:
                        h5_strarr = f[quoted_sentence]
                        output_dict[sentence] = list(h5_strarr.asstr())
                        missing_sentences_set.remove(sentence)

        for text_input, words in output_dict.items():
            output_dict[text_input] = words

        if len(missing_sentences_set) > 0:
            new_output_dict = {}
            missing_sentences = sorted(list(missing_sentences_set))
            words_list = self.lemmatizer.batch_lemmatize(missing_sentences)

            for sentence, words in zip(missing_sentences, words_list):
                for word in words:
                    assert (
                        SEP_CHAR not in word
                    ), f"SEP_CHAR {SEP_CHAR} found in word {word} from sentence {sentence}"
                new_output_dict[sentence] = words

            if self.save_to_db:
                # save lemmas to db
                # single write multi read - use lockfile to make sure only one process writes
                os.makedirs(self.h5_file.parent, exist_ok=True)
                lockfile = self.h5_file.parent / f"{self.h5_file.name}.lock"
                assert (
                    not lockfile.is_file()
                ), f"lockfile {lockfile} exists with content {lockfile.read_text(encoding='utf-8')}"
                lockfile.write_text(f"locked at {datetime.datetime.now()}", encoding="utf-8")

                with h5py.File(self.h5_file, "a", libver="latest") as f:
                    f.swmr_mode = True
                    for i, (sentence, words) in enumerate(new_output_dict.items()):
                        quoted_sentence = quote_with_urlparse(sentence, prefix="q")
                        # words_raw = SEP_CHAR.join(words)
                        if quoted_sentence in f:
                            continue
                        f.create_dataset(quoted_sentence, data=words, dtype=h5py.string_dtype())
                    f.flush()
                lockfile.unlink()
            output_dict.update(new_output_dict)
        output_list = [output_dict[sentence] for sentence in sentences]
        return output_list


def get_lemmatizer(
    lemm_type: str = "spacy",
    lemm_name: str = SPACY_DEFAULT_EN,
    verbose: bool = False,
    use_db: bool = True,
    compute_missing: bool = True,
    save_to_db: bool = True,
    h5_file: Path = None,
) -> LemmatizerInterface:
    """

    Args:
        lemm_type: must be spacy
        lemm_name: name of the lemmatizer
        verbose:
        use_db: cache to h5 file
        compute_missing: compute if missing in cache
        save_to_db: save to db after computing
        h5_file: path to h5 file

    Returns:

    """
    assert lemm_type == "spacy", f"lemm_type {lemm_type} not supported"
    lemmatizer = LemmatizerSpacy(name=lemm_name, verbose=verbose)
    if not use_db:
        return lemmatizer
    lemmatizer_db = LemmatizerDbWrapper(
        lemmatizer=lemmatizer,
        compute_missing=compute_missing,
        save_to_db=save_to_db,
        h5_file=h5_file,
    )
    return lemmatizer_db



import math
import numpy as np
import torch
from typing import Optional, Union, Tuple, List

from packg.maths import np_round_half_down


def torch_stable_softmax(inp, temp=1.0, dim=-1):
    x: torch.Tensor = inp / temp
    max_values, max_indices = x.max(dim=dim, keepdim=True)
    z = x - max_values
    numerator = torch.exp(z)
    denominator = torch.sum(numerator, dim=dim, keepdim=True)
    softmax = numerator / denominator
    return softmax


def torch_cosine_similarity(x: torch.Tensor, y: torch.Tensor, eps: float = 1e-8) -> torch.Tensor:
    """
    Compute cosine similarity between two matrices x and y
    or two vectors x and y.
    """
    eps = torch.tensor(eps, device=x.device)
    x = x / torch.maximum(torch.norm(x, dim=-1, keepdim=True), eps)
    y = y / torch.maximum(torch.norm(y, dim=-1, keepdim=True), eps)
    return torch.matmul(x, y.T)


def compute_indices(
    num_frames_orig: int, num_frames_target: int, is_train: bool = False
) -> np.ndarray:
    """
    Given two sequence lengths n_orig and n_target, sample n_target indices from the range [0, n_orig-1].

    Random sample approximately from intervals during training:
    with factor f = n_orig / n_target, sample in the range [i*f, (i+1)*f].
    Center sample in the same range during validation.

    Args:
        num_frames_orig: Original sequence length n_orig.
        num_frames_target: Target sequence length n_target.
        is_train: enable random sampling of indices. if False, center sample.

    Returns:
        Indices with shape (n_target)
    """
    # random sampling during training
    if is_train:
        # create rounded start points
        start_points = np.linspace(0, num_frames_orig, num_frames_target, endpoint=False)
        start_points = np_round_half_down(start_points).astype(int)

        # compute random offsets s.t. the sum of offsets equals num_frames_orig
        offsets = start_points[1:] - start_points[:-1]
        np.random.shuffle(offsets)
        last_offset = num_frames_orig - np.sum(offsets)
        offsets = np.concatenate([offsets, np.array([last_offset])])

        # compute new start points as cumulative sum of offsets
        new_start_points = np.cumsum(offsets) - offsets[0]

        # move offsets to the left so they fit the new start points
        offsets = np.roll(offsets, -1)

        # now randomly sample in the uniform intervals given by the offsets
        random_offsets = offsets * np.random.rand(num_frames_target)

        # calculate indices and floor them to get ints
        indices = new_start_points + random_offsets
        indices = np.floor(indices).astype(int)
        return indices
    # center sampling during validation
    # compute the linspace and offset it so its centered
    start_points = np.linspace(0, num_frames_orig, num_frames_target, endpoint=False)
    offset = num_frames_orig / num_frames_target / 2
    indices = start_points + offset
    # floor the result to get ints
    indices = np.floor(indices).astype(int)
    return indices


def expand_video_segment(
    num_frames_video: int, min_frames_seg: int, start_frame_seg: int, stop_frame_seg: int
):
    """
    Expand a given video segment defined by start and stop frame to have at least a minimum number of frames.

    Args:
        num_frames_video: Total number of frames in the video.
        min_frames_seg: Target minimum number of frames in the segment.
        start_frame_seg: Current start frame of the segment.
        stop_frame_seg: Current stop frame of the segment.

    Returns:
        Tuple of start frame, stop frame, flag whether the segment was changed.
    """
    num_frames_seg = stop_frame_seg - start_frame_seg
    changes = False
    if min_frames_seg > num_frames_video:
        min_frames_seg = num_frames_video
    if num_frames_seg < min_frames_seg:
        while True:
            if start_frame_seg > 0:
                start_frame_seg -= 1
                num_frames_seg += 1
                changes = True
            if num_frames_seg == min_frames_seg:
                break
            if stop_frame_seg < num_frames_video:
                stop_frame_seg += 1
                num_frames_seg += 1
                changes = True
            if num_frames_seg == min_frames_seg:
                break
    return start_frame_seg, stop_frame_seg, changes


def chunked_mean(arr, max_n=10000):
    ct = 0
    total = 0.0
    for n in range(0, len(arr), max_n):
        n_stop = n + max_n
        arr_sub = arr[n:n_stop]
        ct += len(arr_sub)
        this_sum = np.sum(arr_sub)
        # print(this_sum)
        total += this_sum
    return total / ct


def np_softmax(x: np.ndarray, axis: Optional[Union[int, Tuple[int, ...]]] = -1) -> np.ndarray:
    x_max = np.max(x, axis, keepdims=True)
    unnormalized = np.exp(x - x_max)
    return unnormalized / np.sum(unnormalized, axis, keepdims=True)


def distribute_evenly(num: int, num_chunks: int) -> List[int]:
    """Distribute `num` evenly into `num_chunks` parts.

    Args:
        num: Number to distribute.
        num_chunks: Number of parts to distribute into.

    Returns:
        Array of length `num_chunks` with the distribution.
    """
    step_size = math.floor(num / num_chunks)
    per_shard = [step_size] * num_chunks
    missing = num - sum(per_shard)
    for i in range(missing):
        per_shard[i] += 1
    return per_shard



from __future__ import annotations

import numpy as np
import torch

from packg import Const


class NormsC(Const):
    LINEAR = "linear"


def compress_fp32_to_uint8_numpy(
    feat: np.ndarray, axis: int = -1, norm: str = "linear"
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Compress features into uint8 format for storage.

    Args:
        feat: Input float features.
        axis: Which axis to reduce over.
        norm: Norm to use for scaling: "linear" or "exp".
            Depends on the data, but mostly linear is better.

    Returns:
        feat_uint8: Compressed features, dtype uint8, same shape as input.
        mins: Min values used for scaling, same float dtype as input. same shape as input,
            except the reduction axis has length 1.
        maxs: See mins
    """
    if norm == NormsC.LINEAR:
        mins = np.min(feat, axis=axis, keepdims=True)
        maxs = np.max(feat, axis=axis, keepdims=True)
        # safeguard against div by zero
        problem_idx = np.where((maxs - mins) ** 2 < 1e-16)
        maxs[problem_idx] = mins[problem_idx] + 1e-8
        feat_uint8 = np.round((feat - mins) * 255 / (maxs - mins)).astype(np.uint8)
        return feat_uint8, mins, maxs
    raise ValueError(f"Unknown norm '{norm}', possible values {list(NormsC.values())}")


def decompress_uint8_to_fp32_numpy(
    feat_uint8: np.ndarray, mins: np.ndarray, maxs: np.ndarray, norm: str = "linear"
) -> np.ndarray:
    """
    Undo the compression from compress_fp32_to_uint8_numpy
    """
    if norm == NormsC.LINEAR:
        return feat_uint8.astype(np.float32) / 255.0 * (maxs - mins) + mins
    raise ValueError(f"Unknown norm '{norm}', possible values {list(NormsC.values())}")


def compress_fp32_to_uint8_torch(
    feat: torch.Tensor, axis: int = -1, norm: str = "linear"
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    See compress_fp32_to_uint8_numpy
    """
    if norm == NormsC.LINEAR:
        mins, _mins_idx = torch.min(feat, dim=axis, keepdim=True)
        maxs, _maxs_idx = torch.max(feat, dim=axis, keepdim=True)
        # safeguard against div by zero
        problem_idx = torch.where((maxs - mins) ** 2 < 1e-16)
        maxs[problem_idx] = mins[problem_idx] + 1e-8
        feat_uint8 = torch.round((feat - mins) * 255 / (maxs - mins)).to(torch.uint8)
        return feat_uint8, mins, maxs
    raise ValueError(f"Unknown norm '{norm}', possible values {list(NormsC.values())}")


def decompress_uint8_to_fp32_torch(
    feat_uint8: torch.Tensor, mins: torch.Tensor, maxs: torch.Tensor, norm: str = "linear"
) -> torch.Tensor:
    """
    Undo the compression from compress_fp32_to_uint8_torch
    """
    if norm == NormsC.LINEAR:
        return feat_uint8.float() / 255.0 * (maxs - mins) + mins
    raise ValueError(f"Unknown norm '{norm}', possible values {list(NormsC.values())}")


def decompress_uint8_to_torch(
    feat_uint8: torch.Tensor,
    mins: torch.Tensor,
    maxs: torch.Tensor,
    norm: str = "linear",
    dtype=torch.float32,
) -> torch.Tensor:
    """
    Undo the compression from compress_fp32_to_uint8_torch
    """
    if norm == NormsC.LINEAR:
        return feat_uint8.to(dtype) / 255.0 * (maxs - mins) + mins
    raise ValueError(f"Unknown norm '{norm}', possible values {list(NormsC.values())}")
